import logging

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class SelectorBuilder(ElementSelectorBuilder):
    def _build_wd_selector(self, selectors):
        if any(isinstance(val, Pattern) for val in selectors.values()):
            return None

        expressions = ['./th', './td']
        attr_expr = self.xpath_builder.attribute_expression(None, selectors)

        if attr_expr:
            expressions = ['{}[{}]'.format(e, attr_expr) for e in expressions]

        xpath = " | ".join(expressions)

        logging.debug({'build_wd_selector': xpath})

        return ['xpath', xpath]



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement
from ..user_editable import UserEditable


@six.add_metaclass(MetaHTMLElement)
class TextArea(UserEditable, HTMLElement):
    ATTRIBUTES = ['value']



import six

from .html_elements import HTMLElement
from ..exception import Error
from ..meta_elements import MetaHTMLElement
from ..row_container import RowContainer


@six.add_metaclass(MetaHTMLElement)
class Table(RowContainer, HTMLElement):
    def __iter__(self):
        """
        Yields each TableRow associated with this table
        :rtype: iter

        :Example:

        table = browser.table(id='some_table')
        for row in table:
            print(row.text)
        """
        for row in self.rows():
            yield row

    @property
    def dicts(self):
        """
        Represents table rows as dictionaries
        :rtype: list[dict]
        """
        all_rows = self.rows().locate()
        if len(all_rows) > 0:
            header_row = all_rows[0]
        else:
            raise Error('no rows in table')
        headers = [cell.text for cell in header_row.ths()]
        result = []

        for row in list(all_rows)[1:]:
            self._cell_size_check(header_row, row)
            result.append(dict(zip(headers, [cell.text for cell in row.cells()])))
        return result

    def headers(self, row=None):
        """
        Returns first row of Table with proper subtype
        :rtype: TableCellCollection
        """
        row = row or self.rows()[0]
        if row.th().exist:
            return row.ths()
        else:
            return row.tds()

    def __getitem__(self, idx):
        """
        Returns row of this table with given index

        :param idx: row index
        :rtype: nerodia.elements.row.Row
        """
        return self.row(index=idx)

    # private

    def _cell_size_check(self, header_row, cell_row):
        header_size = len(header_row.cells())
        row_size = len(cell_row.cells())
        if header_size == row_size:
            return

        index = cell_row.selector.get('index')
        row_id = 'row at index {}'.format(index - 1) if index is not None else 'designated row'
        raise Error('{} has {} cells, while header row has '
                    '{}'.format(row_id, row_size, header_size))



import logging

from ..element.selector_builder import SelectorBuilder as ElementSelectorBuilder, \
    XPath as ElementXPath
from ...elements.text_field import TextField
from ...xpath_support import XpathSupport

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class SelectorBuilder(ElementSelectorBuilder):
    # private

    def _build_wd_selector(self, selectors):
        if any(isinstance(val, Pattern) for val in selectors.values()):
            return None

        selectors.pop('tag_name', None)

        input_attr_exp = self.xpath_builder.attribute_expression('input', selectors)

        xpath = './/input[(not(@type) or ({}))'.format(self._negative_type_expr)
        if input_attr_exp:
            xpath += ' and {}'.format(input_attr_exp)
        xpath += ']'

        logging.debug({'build_wd_selector': xpath})

        return ['xpath', xpath]

    @property
    def _negative_type_expr(self):
        types = ['{}!={!r}'.format(XpathSupport.lower('@type'), typ) for
                 typ in TextField.NON_TEXT_TYPES]
        return ' and '.join(types)


class XPath(ElementXPath):
    @staticmethod
    def lhs_for(building, key):
        if building == 'input' and key == 'text':
            return '@value'
        else:
            return super(XPath, XPath).lhs_for(building, key)



from contextlib import contextmanager


class AfterHooks(object):
    """
    After hooks are blocks that run after certain browser events.
    They are generally used to ensure application under test does not encounter
    any error and are automatically executed after following events:
        1. Open URL.
        2. Refresh page.
        3. Click, double-click or right-click on element.
        4. Alert closing.
    """

    def __init__(self, browser):
        self.browser = browser
        self.after_hooks = []

    def add(self, method=None, after_hook=None):
        """
        Adds new after hook
        :param method: callable method
        :param after_hook: callable object

        :Example:

        from __future__ import print_function
        browser.after_hooks.add(method=lambda b: 'Server Error' in browser.text and \
            print('Application exception or 500 error!')
        browser.goto('watir.github.io/404')    #=> 'Application exception or 500 error!'
        """
        if method:
            self.after_hooks.append(method)
        elif callable(after_hook):
            self.after_hooks.append(after_hook)
        else:
            raise ValueError('expected method or callable class')

    def delete(self, after_hook):
        """
        Deletes after hook
        :param after_hook: hook to delete

        :Example:

        from __future__ import print_function
        browser.after_hooks.add(method=lambda b: 'Server Error' in browser.text and \
            print('Application exception or 500 error!')
        browser.goto('watir.github.io/404')    #=> 'Application exception or 500 error!'
        browser.after_hooks.delete(browser.after_hooks[0])
        browser.refresh
        """
        self.after_hooks.remove(after_hook)

    def run(self):
        """ Runs after hooks """
        if self.after_hooks and self.browser.window().present and not self.browser.alert.exists:
            for hook in self.after_hooks:
                hook(self.browser)

    @contextmanager
    def without(self):
        """
        Executes a block without running error after hooks

        :Example:

        with browser.after_hooks.without():
            self.browser.element(name='new_user_button').click()
        """
        current_after_hooks = self.after_hooks
        self.after_hooks = []
        yield
        self.after_hooks = current_after_hooks

    def __len__(self):
        """
        Returns number of after hooks
        :rtype: int

        :Example:

        browser.after_hooks.add(method=lambda b: print('Some after_hook'))
        len(browser.after_hooks)    #=> 1
        """
        return len(self.after_hooks)

    def __getitem__(self, index):
        """
        Gets the after hook at the given index
        :param index: index of the after hook
        :return: callable object
        """
        return self.after_hooks[index]



class UserEditable(object):

    def set(self, *args):
        """
        Clear the element, then type in the given value
        :param args: value to set the input to
        """
        def func():
            self.el.clear()
            self.el.send_keys(*args)
        self._element_call(func, self.wait_for_writable)

    @property
    def value(self):
        return self.attribute_value('value')

    @value.setter  # alias
    def value(self, *args):
        self.set(*args)

    def js_set(self, *args):
        """
        Uses JavaScript to enter most of the given value
        Selenium is used to enter the first and last characters (in order to raise an error if the
        element would normally raise on sending keys)

        :param value: value to set the input to
        """
        from nerodia.exception import Error
        value = ''.join(args)
        if value:
            self.set(value[0])
            if len(value) > 1:
                self._element_call(lambda: self._execute_js('setValue', self.el, value[:-1]))
                self.append(value[-1])
        if self.value != value:
            raise Error('js_set value does not match expected input')

    def append(self, *args):
        """
        Appends the given value to the text in the text field
        :param args: value to add to the input
        """
        self.send_keys(*args)

    def clear(self):
        """
        Clears the text field
        :return:
        """
        self._element_call(lambda: self.el.clear(), self.wait_for_writable)



import six

import nerodia
from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class TableCell(HTMLElement):
    @property
    def column_header(self):
        current_row = self.parent(tag_name='tr')
        return self._header_row(current_row, index=len(self.previous_siblings())).text

    def sibling_from_header(self, **opt):
        current_row = self.parent(tag_name='tr')
        header = self._header_row(current_row, **opt)
        index = len(header.previous_siblings())

        return self.__class__(current_row, {'tag_name': 'td', 'index': index})

    # private

    def _header_row(self, current_row, **opt):
        table = self.parent(tag_name='table')
        header_row = table.tr()

        table._cell_size_check(header_row, current_row)

        header_type = 'th' if table.th().exist else 'tr'
        opt['tag_name'] = header_type

        return nerodia.tag_to_class[header_type](header_row, opt)



import six

from .html_elements import InputCollection
from .input import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Radio(Input):
    def __init__(self, *args, **kwargs):
        super(Radio, self).__init__(*args, **kwargs)
        if 'text' in self.selector:
            self.selector['label'] = self.selector.pop('text')

    @property
    def is_set(self):
        """
        Returns True if the element is selected
        :rtype: bool
        """
        return self._element_call(lambda: self.el.is_selected())

    is_selected = is_set

    def set(self):
        """ Selects the radio input """
        if not self.is_set:
            self.click()

    select = set

    @property
    def text(self):
        lab = self.label()
        return lab.text if lab.exists else ''


@six.add_metaclass(MetaHTMLElement)
class RadioCollection(InputCollection):
    # private

    @property
    def _element_class(self):
        return Radio



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class DList(HTMLElement):
    def to_dict(self):
        keys = [e.text for e in self.dts()]
        values = [e.text for e in self.dds()]
        return dict(zip(keys, values))



import six
from selenium.common.exceptions import NoSuchFrameException

from .html_elements import HTMLElement, HTMLElementCollection
from ..exception import UnknownFrameException
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class IFrame(HTMLElement):
    def locate(self):
        if not self.selector:
            return None
        self.query_scope._ensure_context()

        selector = self.selector.copy()
        selector.update({'tag_name': self._frame_tag})
        element_validator = self._element_validator_class()
        selector_builder = self._selector_builder_class(self.query_scope,
                                                        selector, self.ATTRIBUTES)
        self.locator = self._locator_class(self.query_scope, selector, selector_builder,
                                           element_validator)

        element = self.locator.locate()
        if not element:
            raise self._unknown_exception(
                'unable to locate {} using {}'.format(self.selector['tag_name'],
                                                      self.selector_string))
        self.el = FramedDriver(element, self.browser)
        return self.el

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        self.locate()
        other.locate()
        return self.wd == other.wd.wd if isinstance(other.wd, FramedDriver) else other.wd

    def switch_to(self):
        self.locate().switch()

    def assert_exists(self):
        if self.el and not self.selector:
            raise UnknownFrameException(
                'wrapping a Selenium element as a Frame is not currently supported')
        return super(IFrame, self).assert_exists()

    @property
    def text(self):
        """
        Returns text of iFrame body
        :rtype: str
        """
        return self.body().text

    @property
    def html(self):
        """
        Returns HTML code of iFrame
        :rtype: str
        """
        self.wait_for_exists()
        return self.el.page_source

    def execute_script(self, script, *args):
        """ Executes JavaScript in context of frame """
        from .element import Element
        args = [e.wd if isinstance(e, Element) else e for e in args]
        returned = self.driver.execute_script(script, *args)

        return self.browser._wrap_elements_in(self, returned)

    # private

    def _ensure_context(self):
        self.switch_to()

    @property
    def _frame_tag(self):
        return 'iframe'

    @property
    def _unknown_exception(self):
        return UnknownFrameException


@six.add_metaclass(MetaHTMLElement)
class IFrameCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Frame(IFrame):
    # private

    @property
    def _frame_tag(self):
        return 'frame'


@six.add_metaclass(MetaHTMLElement)
class FrameCollection(IFrameCollection):
    pass


class FramedDriver(object):
    def __init__(self, element, browser):
        self.el = element
        self.browser = browser
        self.driver = browser.driver

    def __eq__(self, other):
        return self.wd == other.wd

    eql = __eq__

    def send_keys(self, *args):
        self.switch()
        self.driver.switch_to.active_element.send_keys(*args)

    @property
    def wd(self):
        return self.el

    def __getattr__(self, meth):
        if hasattr(self.driver, meth):
            self.switch()
            return getattr(self.driver, meth)
        else:
            return getattr(self.el, meth)

    def switch(self):
        try:
            self.driver.switch_to.frame(self.el)
            self.browser.default_context = False
        except NoSuchFrameException as e:
            raise UnknownFrameException(e)



import six

from .element import Element
from ..element_collection import ElementCollection
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class HTMLElement(Element):
    _attr_id = (str, 'id')
    _attr_class_name = (str, 'className')


@six.add_metaclass(MetaHTMLElement)
class HTMLElementCollection(ElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class FontCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Directory(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DirectoryCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class FrameSet(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class FrameSetCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Marquee(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MarqueeCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Applet(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class AppletCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Canvas(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class CanvasCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Template(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TemplateCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Script(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ScriptCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Dialog(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DialogCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class MenuItem(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MenuItemCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Menu(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MenuCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Details(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DetailsCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Legend(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class LegendCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class FieldSet(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class FieldSetCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Meter(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MeterCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Progress(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ProgressCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Output(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class OutputCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Keygen(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class KeygenCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TextAreaCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class OptionCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class OptGroup(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class OptGroupCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class DataList(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DataListCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class SelectCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class ButtonCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class InputCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Label(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class LabelCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class FormCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableCellCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableHeaderCell(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableHeaderCellCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableDataCellCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableRowCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableSectionCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableCol(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableColCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableCaption(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TableCaptionCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class AreaCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Map(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MapCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Media(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MediaCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Audio(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class AudioCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Video(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class VideoCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Track(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TrackCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Param(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ParamCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Object(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ObjectCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Embed(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class EmbedCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class IFrameCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class ImageCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Source(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class SourceCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Picture(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class PictureCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Mod(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ModCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class BR(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class BRCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Span(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class SpanCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Time(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TimeCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Data(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DataCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class AnchorCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Div(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class DivCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class DListCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class LI(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class LICollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class UListCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class OListCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Quote(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class QuoteCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Pre(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class PreCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class HR(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class HRCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Paragraph(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class ParagraphCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Heading(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class HeadingCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Body(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class BodyCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Style(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class StyleCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Meta(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class MetaCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Base(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class BaseCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Title(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class TitleCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Head(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class HeadCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Html(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class HtmlCollection(HTMLElementCollection):
    pass


@six.add_metaclass(MetaHTMLElement)
class Unknown(HTMLElement):
    pass


@six.add_metaclass(MetaHTMLElement)
class UnknownCollection(HTMLElementCollection):
    pass



try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class Validator(object):

    @staticmethod
    def validate(element, selector):
        selector_tag_name = selector.get('tag_name')
        element_tag_name = element.tag_name.lower()

        if selector_tag_name:
            if Validator.match_str_or_regex(selector_tag_name, element_tag_name):
                return element
        else:
            return None

    @staticmethod
    def match_str_or_regex(str_or_regex, term):
        if isinstance(str_or_regex, Pattern) and str_or_regex.search(term):
            return True
        elif str_or_regex == term:
            return True
        else:
            return False



# Autogenerated from HTML specification. Edits may be lost.

htmlelement = [
    (str, 'title', 'title'),
    (str, 'lang', 'lang'),
    (bool, 'translate', 'translate'),
    (str, 'dir', 'dir'),
    (str, 'dataset', 'dataset'),
    (bool, 'hidden', 'hidden'),
    (int, 'tabindex', 'tabIndex'),
    (str, 'accesskey', 'accessKey'),
    (bool, 'draggable', 'draggable'),
    (str, 'dropzone', 'dropzone'),
    (str, 'contextmenu', 'contextMenu'),
    (bool, 'spellcheck', 'spellcheck'),
    (str, 'onabort', 'onabort'),
    (str, 'onblur', 'onblur'),
    (str, 'oncancel', 'oncancel'),
    (str, 'oncanplay', 'oncanplay'),
    (str, 'oncanplaythrough', 'oncanplaythrough'),
    (str, 'onchange', 'onchange'),
    (str, 'onclick', 'onclick'),
    (str, 'onclose', 'onclose'),
    (str, 'oncontextmenu', 'oncontextmenu'),
    (str, 'oncuechange', 'oncuechange'),
    (str, 'ondblclick', 'ondblclick'),
    (str, 'ondrag', 'ondrag'),
    (str, 'ondragend', 'ondragend'),
    (str, 'ondragenter', 'ondragenter'),
    (str, 'ondragexit', 'ondragexit'),
    (str, 'ondragleave', 'ondragleave'),
    (str, 'ondragover', 'ondragover'),
    (str, 'ondragstart', 'ondragstart'),
    (str, 'ondrop', 'ondrop'),
    (str, 'ondurationchange', 'ondurationchange'),
    (str, 'onemptied', 'onemptied'),
    (str, 'onended', 'onended'),
    (str, 'onerror', 'onerror'),
    (str, 'onfocus', 'onfocus'),
    (str, 'oninput', 'oninput'),
    (str, 'oninvalid', 'oninvalid'),
    (str, 'onkeydown', 'onkeydown'),
    (str, 'onkeypress', 'onkeypress'),
    (str, 'onkeyup', 'onkeyup'),
    (str, 'onload', 'onload'),
    (str, 'onloadeddata', 'onloadeddata'),
    (str, 'onloadedmetadata', 'onloadedmetadata'),
    (str, 'onloadstart', 'onloadstart'),
    (str, 'onmousedown', 'onmousedown'),
    (str, 'onmouseenter', 'onmouseenter'),
    (str, 'onmouseleave', 'onmouseleave'),
    (str, 'onmousemove', 'onmousemove'),
    (str, 'onmouseout', 'onmouseout'),
    (str, 'onmouseover', 'onmouseover'),
    (str, 'onmouseup', 'onmouseup'),
    (str, 'onwheel', 'onwheel'),
    (str, 'onpause', 'onpause'),
    (str, 'onplay', 'onplay'),
    (str, 'onplaying', 'onplaying'),
    (str, 'onprogress', 'onprogress'),
    (str, 'onratechange', 'onratechange'),
    (str, 'onreset', 'onreset'),
    (str, 'onresize', 'onresize'),
    (str, 'onscroll', 'onscroll'),
    (str, 'onseeked', 'onseeked'),
    (str, 'onseeking', 'onseeking'),
    (str, 'onselect', 'onselect'),
    (str, 'onshow', 'onshow'),
    (str, 'onstalled', 'onstalled'),
    (str, 'onsubmit', 'onsubmit'),
    (str, 'onsuspend', 'onsuspend'),
    (str, 'ontimeupdate', 'ontimeupdate'),
    (str, 'ontoggle', 'ontoggle'),
    (str, 'onvolumechange', 'onvolumechange'),
    (str, 'onwaiting', 'onwaiting'),
    (str, 'oncopy', 'oncopy'),
    (str, 'oncut', 'oncut'),
    (str, 'onpaste', 'onpaste'),
    (str, 'content_editable', 'contentEditable'),
    (bool, 'content_editable', 'isContentEditable')
]

HTMLElementCollection = []

font = [
    (str, 'color', 'color'),
    (str, 'face', 'face')
]

FontCollection = []

directory = [
    (bool, 'compact', 'compact')
]

DirectoryCollection = []

frameset = [
    (str, 'cols', 'cols'),
    (str, 'onafterprint', 'onafterprint'),
    (str, 'onbeforeprint', 'onbeforeprint'),
    (str, 'onbeforeunload', 'onbeforeunload'),
    (str, 'onhashchange', 'onhashchange'),
    (str, 'onlanguagechange', 'onlanguagechange'),
    (str, 'onmessage', 'onmessage'),
    (str, 'onoffline', 'onoffline'),
    (str, 'ononline', 'ononline'),
    (str, 'onpagehide', 'onpagehide'),
    (str, 'onpageshow', 'onpageshow'),
    (str, 'onrejectionhandled', 'onrejectionhandled'),
    (str, 'onpopstate', 'onpopstate'),
    (str, 'onstorage', 'onstorage'),
    (str, 'onunhandledrejection', 'onunhandledrejection'),
    (str, 'onunload', 'onunload')
]

FrameSetCollection = []

marquee = [
    (str, 'behavior', 'behavior'),
    (str, 'bgcolor', 'bgColor'),
    (str, 'direction', 'direction'),
    (int, 'hspace', 'hspace'),
    (int, 'loop', 'loop'),
    (int, 'scrollamount', 'scrollAmount'),
    (int, 'scrolldelay', 'scrollDelay'),
    (bool, 'truespeed', 'trueSpeed'),
    (int, 'vspace', 'vspace'),
    (str, 'onbounce', 'onbounce'),
    (str, 'onfinish', 'onfinish'),
    (str, 'onstart', 'onstart')
]

MarqueeCollection = []

applet = [
    (str, 'align', 'align'),
    (str, 'alt', 'alt'),
    (str, 'archive', 'archive'),
    (str, 'code', 'code'),
    (str, 'codebase', 'codeBase'),
    (int, 'hspace', 'hspace'),
    (str, 'name', 'name'),
    (str, 'object', 'object'),
    (int, 'vspace', 'vspace')
]

AppletCollection = []

Canvas = []

CanvasCollection = []

template = [
    (str, 'content', 'content')
]

TemplateCollection = []

script = [
    (str, 'src', 'src'),
    (str, 'type', 'type'),
    (str, 'charset', 'charset'),
    (bool, 'async', 'async'),
    (bool, 'defer', 'defer'),
    (str, 'crossorigin', 'crossOrigin'),
    (str, 'nonce', 'nonce'),
    (str, 'event', 'event'),
    (str, 'for', 'htmlFor')
]

ScriptCollection = []

menuitem = [
    (str, 'type', 'type'),
    (str, 'label', 'label'),
    (str, 'icon', 'icon'),
    (bool, 'disabled', 'disabled'),
    (bool, 'checked', 'checked'),
    (str, 'radiogroup', 'radiogroup'),
    (bool, 'default', 'default')
]

MenuItemCollection = []

menu = [
    (str, 'type', 'type'),
    (str, 'label', 'label'),
    (bool, 'compact', 'compact')
]

MenuCollection = []

details = [
    (bool, 'open', 'open')
]

DetailsCollection = []

legend = [
    (str, 'form', 'form'),
    (str, 'align', 'align')
]

LegendCollection = []

fieldset = [
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'name', 'name'),
    (str, 'type', 'type'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage')
]

FieldSetCollection = []

meter = [
    (float, 'value', 'value'),
    (float, 'min', 'min'),
    (float, 'max', 'max'),
    (float, 'low', 'low'),
    (float, 'high', 'high'),
    (float, 'optimum', 'optimum'),
    (str, 'labels', 'labels')
]

MeterCollection = []

progress = [
    (float, 'value', 'value'),
    (float, 'max', 'max'),
    (float, 'position', 'position'),
    (str, 'labels', 'labels')
]

ProgressCollection = []

output = [
    (str, 'for', 'htmlFor'),
    (str, 'form', 'form'),
    (str, 'name', 'name'),
    (str, 'type', 'type'),
    (str, 'defaultvalue', 'defaultValue'),
    (str, 'value', 'value'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels')
]

OutputCollection = []

keygen = [
    (bool, 'autofocus', 'autofocus'),
    (str, 'challenge', 'challenge'),
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'keytype', 'keytype'),
    (str, 'name', 'name'),
    (str, 'type', 'type'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels')
]

KeygenCollection = []

textarea = [
    (str, 'autocomplete', 'autocomplete'),
    (bool, 'autofocus', 'autofocus'),
    (int, 'cols', 'cols'),
    (str, 'dirname', 'dirName'),
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'inputmode', 'inputMode'),
    (int, 'maxlength', 'maxLength'),
    (int, 'minlength', 'minLength'),
    (str, 'name', 'name'),
    (str, 'placeholder', 'placeholder'),
    (bool, 'readonly', 'readOnly'),
    (bool, 'required', 'required'),
    (str, 'wrap', 'wrap'),
    (str, 'type', 'type'),
    (str, 'defaultvalue', 'defaultValue'),
    (int, 'textlength', 'textLength'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels'),
    (int, 'selectionstart', 'selectionStart'),
    (int, 'selectionend', 'selectionEnd'),
    (str, 'selectiondirection', 'selectionDirection')
]

TextAreaCollection = []

option = [
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'label', 'label'),
    (bool, 'defaultselected', 'defaultSelected'),
    (bool, 'selected', 'selected'),
    (str, 'value', 'value'),
    (int, 'index', 'index')
]

OptionCollection = []

optgroup = [
    (bool, 'disabled', 'disabled'),
    (str, 'label', 'label')
]

OptGroupCollection = []

datalist = [
    (str, 'options', 'options')
]

DataListCollection = []

select = [
    (str, 'autocomplete', 'autocomplete'),
    (bool, 'autofocus', 'autofocus'),
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (bool, 'multiple', 'multiple'),
    (str, 'name', 'name'),
    (bool, 'required', 'required'),
    (str, 'type', 'type'),
    (str, 'element_options', 'options'),
    (int, 'length', 'length'),
    (str, 'selectedoptions', 'selectedOptions'),
    (int, 'selectedindex', 'selectedIndex'),
    (str, 'value', 'value'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels')
]

SelectCollection = []

button = [
    (bool, 'autofocus', 'autofocus'),
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'formaction', 'formAction'),
    (str, 'formenctype', 'formEnctype'),
    (str, 'formmethod', 'formMethod'),
    (bool, 'formnovalidate', 'formNoValidate'),
    (str, 'formtarget', 'formTarget'),
    (str, 'name', 'name'),
    (str, 'type', 'type'),
    (str, 'value', 'value'),
    (str, 'menu', 'menu'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels')
]

ButtonCollection = []

input = [
    (str, 'accept', 'accept'),
    (str, 'alt', 'alt'),
    (str, 'autocomplete', 'autocomplete'),
    (bool, 'autofocus', 'autofocus'),
    (bool, 'defaultchecked', 'defaultChecked'),
    (bool, 'checked', 'checked'),
    (str, 'dirname', 'dirName'),
    (bool, 'disabled', 'disabled'),
    (str, 'form', 'form'),
    (str, 'files', 'files'),
    (str, 'formaction', 'formAction'),
    (str, 'formenctype', 'formEnctype'),
    (str, 'formmethod', 'formMethod'),
    (bool, 'formnovalidate', 'formNoValidate'),
    (str, 'formtarget', 'formTarget'),
    (bool, 'indeterminate', 'indeterminate'),
    (str, 'inputmode', 'inputMode'),
    (str, 'list', 'list'),
    (str, 'max', 'max'),
    (int, 'maxlength', 'maxLength'),
    (str, 'min', 'min'),
    (int, 'minlength', 'minLength'),
    (bool, 'multiple', 'multiple'),
    (str, 'name', 'name'),
    (str, 'pattern', 'pattern'),
    (str, 'placeholder', 'placeholder'),
    (bool, 'readonly', 'readOnly'),
    (bool, 'required', 'required'),
    (str, 'src', 'src'),
    (str, 'step', 'step'),
    (str, 'type', 'type'),
    (str, 'defaultvalue', 'defaultValue'),
    (str, 'value', 'value'),
    (str, 'valueasdate', 'valueAsDate'),
    (float, 'valueasnumber', 'valueAsNumber'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'labels', 'labels'),
    (int, 'selectionstart', 'selectionStart'),
    (int, 'selectionend', 'selectionEnd'),
    (str, 'selectiondirection', 'selectionDirection'),
    (str, 'align', 'align'),
    (str, 'usemap', 'useMap')
]

InputCollection = []

label = [
    (str, 'form', 'form'),
    (str, 'for', 'htmlFor'),
    (str, 'control', 'control')
]

LabelCollection = []

form = [
    (str, 'accept_charset', 'acceptCharset'),
    (str, 'action', 'action'),
    (str, 'autocomplete', 'autocomplete'),
    (str, 'enctype', 'enctype'),
    (str, 'encoding', 'encoding'),
    (str, 'method', 'method'),
    (str, 'name', 'name'),
    (bool, 'novalidate', 'noValidate'),
    (str, 'target', 'target'),
    (int, 'length', 'length')
]

FormCollection = []

tablecell = [
    (int, 'colspan', 'colSpan'),
    (int, 'rowspan', 'rowSpan'),
    (str, 'headers', 'headers'),
    (int, 'cellindex', 'cellIndex'),
    (str, 'align', 'align'),
    (str, 'axis', 'axis'),
    (str, 'ch', 'ch'),
    (str, 'choff', 'chOff'),
    (bool, 'nowrap', 'noWrap'),
    (str, 'valign', 'vAlign'),
    (str, 'bgcolor', 'bgColor')
]

TableCellCollection = []

tableheadercell = [
    (str, 'scope', 'scope'),
    (str, 'abbr', 'abbr')
]

TableHeaderCellCollection = []

tabledatacell = [
    (str, 'abbr', 'abbr')
]

TableDataCellCollection = []

tablerow = [
    (int, 'rowindex', 'rowIndex'),
    (int, 'sectionrowindex', 'sectionRowIndex'),
    (str, 'align', 'align'),
    (str, 'ch', 'ch'),
    (str, 'choff', 'chOff'),
    (str, 'valign', 'vAlign'),
    (str, 'bgcolor', 'bgColor')
]

TableRowCollection = []

tablesection = [
    (str, 'align', 'align'),
    (str, 'ch', 'ch'),
    (str, 'choff', 'chOff'),
    (str, 'valign', 'vAlign')
]

TableSectionCollection = []

tablecol = [
    (str, 'align', 'align'),
    (str, 'ch', 'ch'),
    (str, 'choff', 'chOff'),
    (str, 'valign', 'vAlign')
]

TableColCollection = []

tablecaption = [
    (str, 'align', 'align')
]

TableCaptionCollection = []

table = [
    (str, 'caption', 'caption'),
    (str, 'tbodies', 'tBodies'),
    (str, 'align', 'align'),
    (str, 'border', 'border'),
    (str, 'frame', 'frame'),
    (str, 'rules', 'rules'),
    (str, 'summary', 'summary'),
    (str, 'bgcolor', 'bgColor'),
    (str, 'cellpadding', 'cellPadding'),
    (str, 'cellspacing', 'cellSpacing')
]

TableCollection = []

area = [
    (str, 'alt', 'alt'),
    (str, 'coords', 'coords'),
    (str, 'shape', 'shape'),
    (str, 'target', 'target'),
    (str, 'download', 'download'),
    (str, 'rel', 'rel'),
    (str, 'rellist', 'relList'),
    (str, 'hreflang', 'hreflang'),
    (str, 'type', 'type'),
    (bool, 'nohref', 'noHref'),
    (str, 'href', 'href'),
    (str, 'origin', 'origin'),
    (str, 'protocol', 'protocol'),
    (str, 'username', 'username'),
    (str, 'password', 'password'),
    (str, 'host', 'host'),
    (str, 'hostname', 'hostname'),
    (str, 'port', 'port'),
    (str, 'pathname', 'pathname'),
    (str, 'search', 'search')
]

AreaCollection = []

map = [
    (str, 'name', 'name'),
    (str, 'areas', 'areas'),
    (str, 'images', 'images')
]

MapCollection = []

media = [
    (str, 'error', 'error'),
    (str, 'src', 'src'),
    (str, 'srcobject', 'srcObject'),
    (str, 'currentsrc', 'currentSrc'),
    (str, 'crossorigin', 'crossOrigin'),
    (int, 'networkstate', 'networkState'),
    (str, 'preload', 'preload'),
    (str, 'buffered', 'buffered'),
    (int, 'readystate', 'readyState'),
    (bool, 'seeking', 'seeking'),
    (float, 'currenttime', 'currentTime'),
    (float, 'duration', 'duration'),
    (bool, 'paused', 'paused'),
    (float, 'defaultplaybackrate', 'defaultPlaybackRate'),
    (float, 'playbackrate', 'playbackRate'),
    (str, 'played', 'played'),
    (str, 'seekable', 'seekable'),
    (bool, 'ended', 'ended'),
    (bool, 'autoplay', 'autoplay'),
    (bool, 'loop', 'loop'),
    (bool, 'controls', 'controls'),
    (float, 'volume', 'volume'),
    (bool, 'muted', 'muted'),
    (bool, 'defaultmuted', 'defaultMuted'),
    (str, 'audiotracks', 'audioTracks'),
    (str, 'videotracks', 'videoTracks'),
    (str, 'texttracks', 'textTracks')
]

MediaCollection = []

Audio = []

AudioCollection = []

video = [
    (int, 'videowidth', 'videoWidth'),
    (int, 'videoheight', 'videoHeight'),
    (str, 'poster', 'poster')
]

VideoCollection = []

track = [
    (str, 'kind', 'kind'),
    (str, 'src', 'src'),
    (str, 'srclang', 'srclang'),
    (str, 'label', 'label'),
    (bool, 'default', 'default'),
    (int, 'readystate', 'readyState'),
    (str, 'track', 'track')
]

TrackCollection = []

param = [
    (str, 'name', 'name'),
    (str, 'value', 'value'),
    (str, 'type', 'type'),
    (str, 'valuetype', 'valueType')
]

ParamCollection = []

object = [
    (str, 'data', 'data'),
    (str, 'type', 'type'),
    (bool, 'typemustmatch', 'typeMustMatch'),
    (str, 'name', 'name'),
    (str, 'form', 'form'),
    (str, 'contentdocument', 'contentDocument'),
    (str, 'contentwindow', 'contentWindow'),
    (bool, 'willvalidate', 'willValidate'),
    (str, 'validity', 'validity'),
    (str, 'validationmessage', 'validationMessage'),
    (str, 'align', 'align'),
    (str, 'archive', 'archive'),
    (str, 'code', 'code'),
    (bool, 'declare', 'declare'),
    (int, 'hspace', 'hspace'),
    (str, 'standby', 'standby'),
    (int, 'vspace', 'vspace'),
    (str, 'codebase', 'codeBase'),
    (str, 'codetype', 'codeType'),
    (str, 'usemap', 'useMap'),
    (str, 'border', 'border')
]

ObjectCollection = []

embed = [
    (str, 'src', 'src'),
    (str, 'type', 'type'),
    (str, 'align', 'align'),
    (str, 'name', 'name')
]

EmbedCollection = []

iframe = [
    (str, 'src', 'src'),
    (str, 'srcdoc', 'srcdoc'),
    (str, 'name', 'name'),
    (str, 'sandbox', 'sandbox'),
    (bool, 'allowfullscreen', 'allowFullscreen'),
    (str, 'contentdocument', 'contentDocument'),
    (str, 'contentwindow', 'contentWindow'),
    (str, 'align', 'align'),
    (str, 'scrolling', 'scrolling'),
    (str, 'frameborder', 'frameBorder'),
    (str, 'marginheight', 'marginHeight'),
    (str, 'marginwidth', 'marginWidth')
]

IFrameCollection = []

image = [
    (str, 'alt', 'alt'),
    (str, 'src', 'src'),
    (str, 'srcset', 'srcset'),
    (str, 'sizes', 'sizes'),
    (str, 'crossorigin', 'crossOrigin'),
    (str, 'usemap', 'useMap'),
    (bool, 'ismap', 'isMap'),
    (int, 'naturalwidth', 'naturalWidth'),
    (int, 'naturalheight', 'naturalHeight'),
    (bool, 'complete', 'complete'),
    (str, 'currentsrc', 'currentSrc'),
    (str, 'name', 'name'),
    (str, 'lowsrc', 'lowsrc'),
    (str, 'align', 'align'),
    (int, 'hspace', 'hspace'),
    (int, 'vspace', 'vspace'),
    (str, 'border', 'border')
]

ImageCollection = []

source = [
    (str, 'srcset', 'srcset'),
    (str, 'sizes', 'sizes'),
    (str, 'media', 'media'),
    (str, 'src', 'src'),
    (str, 'type', 'type')
]

SourceCollection = []

Picture = []

PictureCollection = []

mod = [
    (str, 'cite', 'cite'),
    (str, 'datetime', 'dateTime')
]

ModCollection = []

br = [
    (str, 'clear', 'clear')
]

BRCollection = []

Span = []

SpanCollection = []

time = [
    (str, 'datetime', 'dateTime')
]

TimeCollection = []

data = [
    (str, 'value', 'value')
]

DataCollection = []

anchor = [
    (str, 'target', 'target'),
    (str, 'download', 'download'),
    (str, 'rel', 'rel'),
    (str, 'rev', 'rev'),
    (str, 'rellist', 'relList'),
    (str, 'hreflang', 'hreflang'),
    (str, 'type', 'type'),
    (str, 'coords', 'coords'),
    (str, 'charset', 'charset'),
    (str, 'name', 'name'),
    (str, 'shape', 'shape'),
    (str, 'href', 'href'),
    (str, 'origin', 'origin'),
    (str, 'protocol', 'protocol'),
    (str, 'username', 'username'),
    (str, 'password', 'password'),
    (str, 'host', 'host'),
    (str, 'hostname', 'hostname'),
    (str, 'port', 'port'),
    (str, 'pathname', 'pathname'),
    (str, 'search', 'search')
]

AnchorCollection = []

div = [
    (str, 'align', 'align')
]

DivCollection = []

dlist = [
    (bool, 'compact', 'compact')
]

DListCollection = []

li = [
    (int, 'value', 'value'),
    (str, 'type', 'type')
]

LICollection = []

ulist = [
    (bool, 'compact', 'compact'),
    (str, 'type', 'type')
]

UListCollection = []

olist = [
    (bool, 'reversed', 'reversed'),
    (int, 'start', 'start'),
    (str, 'type', 'type'),
    (bool, 'compact', 'compact')
]

OListCollection = []

quote = [
    (str, 'cite', 'cite')
]

QuoteCollection = []

Pre = []

PreCollection = []

hr = [
    (str, 'align', 'align'),
    (str, 'color', 'color'),
    (bool, 'noshade', 'noShade')
]

HRCollection = []

paragraph = [
    (str, 'align', 'align')
]

ParagraphCollection = []

heading = [
    (str, 'align', 'align')
]

HeadingCollection = []

body = [
    (str, 'link', 'link'),
    (str, 'vlink', 'vLink'),
    (str, 'alink', 'aLink'),
    (str, 'bgcolor', 'bgColor'),
    (str, 'background', 'background'),
    (str, 'onafterprint', 'onafterprint'),
    (str, 'onbeforeprint', 'onbeforeprint'),
    (str, 'onbeforeunload', 'onbeforeunload'),
    (str, 'onhashchange', 'onhashchange'),
    (str, 'onlanguagechange', 'onlanguagechange'),
    (str, 'onmessage', 'onmessage'),
    (str, 'onoffline', 'onoffline'),
    (str, 'ononline', 'ononline'),
    (str, 'onpagehide', 'onpagehide'),
    (str, 'onpageshow', 'onpageshow'),
    (str, 'onrejectionhandled', 'onrejectionhandled'),
    (str, 'onpopstate', 'onpopstate'),
    (str, 'onstorage', 'onstorage'),
    (str, 'onunhandledrejection', 'onunhandledrejection'),
    (str, 'onunload', 'onunload')
]

BodyCollection = []

style = [
    (str, 'media', 'media'),
    (str, 'nonce', 'nonce'),
    (str, 'type', 'type')
]

StyleCollection = []

meta = [
    (str, 'name', 'name'),
    (str, 'http_equiv', 'httpEquiv'),
    (str, 'content', 'content'),
    (str, 'scheme', 'scheme')
]

MetaCollection = []

base = [
    (str, 'href', 'href'),
    (str, 'target', 'target')
]

BaseCollection = []

Title = []

TitleCollection = []

Head = []

HeadCollection = []

html = [
    (str, 'version', 'version')
]

HtmlCollection = []

Unknown = []

UnknownCollection = []



import re
from copy import copy
from itertools import islice
from time import sleep

import six
from selenium.common.exceptions import NoSuchElementException, StaleElementReferenceException
from selenium.webdriver.common.by import By

import nerodia
from .validator import Validator
from ...exception import Error
from ...xpath_support import XpathSupport

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


class Locator(object):
    W3C_FINDERS = {
        'css': By.CSS_SELECTOR,
        'link': By.LINK_TEXT,
        'link_text': By.LINK_TEXT,
        'partial_link_text': By.PARTIAL_LINK_TEXT,
        'tag_name': By.TAG_NAME,
        'xpath': By.XPATH
    }

    # Regular expressions that can be reliably converted to xpath `contains`
    # expressions in order to optimize the locator.
    CONVERTABLE_REGEXP = re.compile(r'\A'
                                    r'([^\[\]\\^$.|?*+()]*)'  # leading literal characters
                                    r'[^|]*?'  # do not try to convert expressions with alternates
                                    r'([^\[\]\\^$.|?*+()]*)'  # trailing literal characters
                                    r'\Z',
                                    re.X)

    def __init__(self, query_scope, selector, selector_builder, element_validator):
        self.query_scope = query_scope  # either element or browser
        self.selector = copy(selector)
        self.selector_builder = selector_builder
        self.element_validator = element_validator
        self.filter_selector = None
        self.normalized_selector = None
        self.driver_scope = None

    def locate(self):
        try:
            elements = self._using_selenium('first')
            return elements if elements else self._using_nerodia('first')
        except (NoSuchElementException, StaleElementReferenceException):
            return None

    def locate_all(self):
        if 'element' in self.selector:
            return [self.selector.get('element')]

        elements = self._using_selenium('all')
        return elements if elements else self._using_nerodia('all')

    # private

    def _using_selenium(self, filter='first'):
        tag = self.selector.pop('tag_name', None)
        if len(self.selector) > 1:
            return

        how = list(self.selector.keys())[0] if self.selector else 'tag_name'
        what = list(self.selector.values())[0] if self.selector else tag

        if not self._wd_is_supported(how, what, tag):
            return

        if filter == 'all':
            return self._locate_elements(how, what)
        else:
            return self._locate_element(how, what)

    def _using_nerodia(self, filter='first'):
        self._create_normalized_selector(filter)
        if self.normalized_selector is None:
            return

        self._create_filter_selector()

        built_selector = self.selector_builder.build(self.normalized_selector.copy())
        if not built_selector:
            raise Error('internal error: unable to build Selenium selector from '
                        '{}'.format(self.normalized_selector))

        how, what = built_selector
        if how == 'xpath':
            what = self._add_regexp_predicates(what)

        if filter == 'all' or len(self.filter_selector) > 0:
            retries = 0
            while True:
                try:
                    elements = self._locate_elements(how, what, self.driver_scope) or []
                    return self._filter_elements(elements, filter=filter)
                except StaleElementReferenceException:
                    retries += 1
                    sleep(0.5)
                    if retries < 2:
                        continue
                    target = 'element collection' if filter == 'all' else 'element'
                    raise Error('Unable to locate {} from {} due to changing '
                                'page'.format(target, self.selector))
        else:
            return self._locate_element(how, what, self.driver_scope)

    def _validate(self, elements, tag_name):
        return all(self.element_validator.validate(el, {'tag_name': tag_name})
                   for el in elements if el is not None)

    def _fetch_value(self, element, how):
        if how == 'text':
            return element.text
        elif how == 'visible':
            return element.is_displayed()
        elif how == 'visible_text':
            return element.text
        elif how == 'tag_name':
            return element.tag_name.lower()
        elif how == 'href':
            href = element.get_attribute('href')
            return href and href.strip()
        else:
            return element.get_attribute(how.replace('_', '-')) or ''

    def _filter_elements(self, elements, filter='first'):
        selector = self.filter_selector.copy()
        if filter == 'first':
            idx = selector.pop('index', 0)
            if idx < 0:
                elements.reverse()
                idx = abs(idx) - 1
            # Generator + slice to avoid fetching values for elements that will be discarded
            matches = (el for el in elements if self._matches_selector(el, selector))
            try:
                return list(islice(matches, idx + 1))[idx]
            except IndexError:
                return None
        else:
            return [el for el in elements if self._matches_selector(el, selector)]

    def _create_normalized_selector(self, filter):
        if self.normalized_selector is not None:
            return self.normalized_selector
        self.driver_scope = self._ensure_scope_context()

        self.normalized_selector = self.selector_builder.normalized_selector

        label_key = None
        if 'label' in self.normalized_selector:
            label_key = 'label'
        elif 'visible_label' in self.normalized_selector:
            label_key = 'visible_label'

        if label_key:
            self._process_label(label_key)
            if self.normalized_selector is None:
                return None

        if 'index' in self.normalized_selector and filter == 'all':
            raise ValueError("can't locate all elements by index")
        return self.normalized_selector

    def _create_filter_selector(self):
        if self.filter_selector:
            return self.filter_selector
        self.filter_selector = {}

        # Remove selectors that can never be used in XPath builder
        for how in ('visible', 'visible_text'):
            if how in self.normalized_selector:
                self.filter_selector[how] = self.normalized_selector.pop(how, None)

        if self._tag_vaildation_required(self.normalized_selector):
            self.filter_selector['tag_name'] = self.normalized_selector.get('tag_name')

        # Regexp locators currently need to be validated even if they are included in the XPath
        # builder
        # TODO: Identify Regexp that can have an exact equivalent using XPath contains (ie would
        # not require filtering) vs approximations (ie would still require filtering)
        for how, what in self.normalized_selector.copy().items():
            if isinstance(what, Pattern):
                self.filter_selector[how] = self.normalized_selector.pop(how, None)

        if self.normalized_selector.get('index') is not None and \
                self.normalized_selector.get('adjacent') is None:
            idx = self.normalized_selector.pop('index')

            # Do not add {'index': 0} filter if the only filter. This will allow using #find_element
            # instead of #find_elements
            implicit_idx_filter = not self.filter_selector and idx == 0
            if not implicit_idx_filter:
                self.filter_selector['index'] = idx

        return self.filter_selector

    def _process_label(self, label_key):
        regexp = isinstance(self.normalized_selector.get(label_key), Pattern)

        if (regexp or label_key == 'visible_label') and \
                self.selector_builder.should_use_label_element:

            label = self._label_from_text(label_key)
            if not label:
                self.normalized_selector = None
                return None

            _id = label.get_attribute('for')
            if _id:
                self.normalized_selector['id'] = _id
            else:
                self.driver_scope = label

    def _label_from_text(self, label_key):
        # TODO: this won't work correctly if @wd is a sub-element
        # TODO: figure out how to do this with find_element
        label_text = self.normalized_selector.pop(label_key, None)
        locator_key = label_key.replace('label', 'text')
        elements = self._locate_elements('tag_name', 'label', self.driver_scope)
        return next((e for e in elements if self._matches_selector(e, {locator_key: label_text})),
                    None)

    def _matches_selector(self, element, selector):
        def check_match(how, what):
            if how == 'tag_name' and isinstance(what, six.string_types):
                return self.element_validator.validate(element, {'tag_name': what})
            else:
                return Validator.match_str_or_regex(what, self._fetch_value(element, how))

        matches = all(check_match(how, what) for how, what in selector.items())

        text_selector = selector.get('text')
        if text_selector is not None:
            from nerodia.elements.element import Element
            text_content = Element(self.query_scope, {'element': element}).\
                _execute_js('getTextContent', element).strip()
            text_content_matches = Validator.match_str_or_regex(text_selector, text_content)
            if matches != text_content_matches:
                key = 'text' if 'text' in self.selector else 'label'
                nerodia.logger.deprecate('Using {!r} locator with RegExp: {!r} to match an element '
                                         'that includes hidden '
                                         'text'.format(key, text_selector.pattern),
                                         'visible_{}'.format(key),
                                         ids=['visible_text'])

        return matches

    @property
    def _can_convert_regexp_to_contains(self):
        return True

    def _add_regexp_predicates(self, what):
        if not self._can_convert_regexp_to_contains:
            return what

        for key, value in self.filter_selector.items():
            if key in ['tag_name', 'text', 'visible_text', 'visible', 'index']:
                continue

            predicates = self._regexp_selector_to_predicates(key, value)
            if predicates:
                what = "({})[{}]".format(what, ' and '.join(predicates))

        return what

    def _regexp_selector_to_predicates(self, key, regex):
        if regex.flags & re.IGNORECASE:
            return []

        match = self.CONVERTABLE_REGEXP.search(regex.pattern)
        if match is None:
            return None

        lhs = self.selector_builder.xpath_builder.lhs_for(None, key)

        return ['contains({}, {})'.format(lhs, XpathSupport.escape(group)) for
                group in match.groups() if group]

    def _tag_vaildation_required(self, selector):
        return any(x in selector for x in ('css', 'xpath')) and 'tag_name' in selector

    def _ensure_scope_context(self):
        return self.query_scope.wd

    def _locate_element(self, how, what, scope=None):
        scope = scope or self.query_scope.wd
        return scope.find_element(self._wd_finder(how), what)

    def _locate_elements(self, how, what, scope=None):
        scope = scope or self.query_scope.wd
        return scope.find_elements(self._wd_finder(how), what)

    def _wd_finder(self, how):
        return self.W3C_FINDERS.get(how, how)

    def _wd_is_supported(self, how, what, tag):
        if how not in self.W3C_FINDERS:
            return False
        if type(what) not in nerodia._str_types:
            return False

        if how in ['partial_link_text', 'link_text', 'link']:
            nerodia.logger.deprecate('{} locator'.format(how), 'visible_text')
            if tag in ['link', None]:
                return True
            raise Exception('Can not use {} locator to find a {} element'.format(how, what))
        elif how == 'tag_name':
            return True
        else:
            if tag:
                return False
        return True



import logging
import re
from importlib import import_module

import six
from selenium.webdriver.common.by import By

import nerodia
from ...xpath_support import XpathSupport

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern

STRING_TYPES = [six.text_type, six.binary_type]


class SelectorBuilder(object):
    VALID_WHATS = [list, Pattern, bool] + STRING_TYPES
    WILDCARD_ATTRIBUTE = re.compile(r'^(aria|data)_(.+)$')

    def __init__(self, query_scope, selector, valid_attributes):
        self.query_scope = query_scope  # either element or browser
        self.selector = selector
        self.valid_attributes = valid_attributes
        self.custom_attributes = []

    @property
    def normalized_selector(self):
        selector = {}

        for how, what in self.selector.items():
            self.check_type(how, what)

            how, what = self._normalize_selector(how, what)
            selector[how] = what

        return selector

    def check_type(self, how, what):
        if how == 'index':
            if not isinstance(what, int):
                raise TypeError('expected {}, got {!r}:{}'.format(int, what, what.__class__))
        elif how == 'visible':
            if not isinstance(what, bool):
                raise TypeError('expected {}, got {!r}:{}'.format(bool, what, what.__class__))
        elif how == 'visible_text':
            if type(what) not in [six.text_type, six.binary_type, Pattern]:
                raise TypeError('expected str or regexp, got {}')
        else:
            if isinstance(what, list) and how != 'class_name':
                raise TypeError("only 'class_name' locator can have a value of a list")
            if type(what) not in self.VALID_WHATS:
                raise TypeError(
                    'expected one of {}, got {!r}:{}'.format(self.VALID_WHATS, what, what.__class__))

    @property
    def should_use_label_element(self):
        return not self._is_valid_attribute('label')

    def build(self, selector):
        if 'xpath' in selector or 'css' in selector:
            return self._given_xpath_or_css(selector)
        built = self._build_wd_selector(selector)
        nerodia.logger.debug('Converted {} to {}'.format(selector, built))
        return built

    @property
    def xpath_builder(self):
        return self._xpath_builder_class(self.should_use_label_element)

    # private

    def _normalize_selector(self, how, what):
        if how in ['tag_name', 'text', 'xpath', 'index', 'class', 'label', 'css', 'visible',
                   'visible_text', 'adjacent']:
            # include 'class' since the valid attribute is 'class_name'
            return [how, what]
        elif how == 'class_name':
            return ['class', what]
        elif how == 'caption':
            return ['text', what]
        else:
            self._check_custom_attribute(how)
            return [how, what]

    def _check_custom_attribute(self, attribute):
        if self._is_valid_attribute(attribute) or self.WILDCARD_ATTRIBUTE.search(attribute):
            return None
        self.custom_attributes.append(attribute)

    def _given_xpath_or_css(self, selector):
        xpath = selector.pop('xpath', None)
        css = selector.pop('css', None)

        if not (xpath or css):
            return None

        if xpath and css:
            raise ValueError("'xpath' and 'css' cannot be combined ({})".format(selector))

        how, what = [None] * 2
        if xpath:
            how = By.XPATH
            what = xpath
        elif css:
            how = By.CSS_SELECTOR
            what = css

        if selector and not self._can_be_combined_with_xpath_or_css(selector):
            raise ValueError('{} cannot be combined with other selectors {})'.format(how, selector))

        return [how, what]

    def _build_wd_selector(self, selectors):
        if any(isinstance(val, Pattern) for val in selectors.values()):
            return None
        return self._build_xpath(selectors)

    def _is_valid_attribute(self, attribute):
        return self.valid_attributes and attribute in self.valid_attributes

    @staticmethod
    def _can_be_combined_with_xpath_or_css(selector):
        keys = list(selector)
        if keys == ['tag_name']:
            return True

        if selector.get('tag_name') == 'input':
            keys.sort()
            return keys == ['tag_name', 'type']
        return False

    def _build_xpath(self, selectors):
        return self.xpath_builder.build(selectors)

    @property
    def _xpath_builder_class(self):
        try:
            mod = import_module(self.__module__)
            return getattr(mod, 'XPath', XPath)
        except ImportError:
            return XPath


class XPath(object):
    def __init__(self, should_use_label_element):
        self.should_use_label_element = should_use_label_element

    def build(self, selectors):
        adjacent = selectors.pop('adjacent', None)
        xpath = self._process_adjacent(adjacent) if adjacent else './/'

        xpath += selectors.pop('tag_name', '*')

        index = selectors.pop('index', None)

        # the remaining entries should be attributes
        if selectors:
            xpath += '[{}]'.format(self.attribute_expression(None, selectors))

        if adjacent and index is not None:
            xpath += "[{}]".format(index + 1)

        logging.debug({'xpath': xpath, 'selectors': selectors})

        return [By.XPATH, xpath]

    # TODO: Get rid of building
    def attribute_expression(self, building, selectors):
        expressions = []
        for key, val in selectors.items():
            if isinstance(val, list) and key == 'class':
                term = '({})'.format(' and '.join([self._build_class_match(v) for v in val]))
            elif isinstance(val, list):
                term = '({})'.format(' or '.join([self.equal_pair(building, key, v) for v in val]))
            elif val is True:
                term = self._attribute_presence(key)
            elif val is False:
                term = self._attribute_absence(key)
            else:
                term = self.equal_pair(building, key, val)
            expressions.append(term)
        return ' and '.join(expressions)

    # TODO: Get rid of building
    def equal_pair(self, building, key, value):
        if key in ['class', 'class_name']:
            if ' ' in value.strip():
                nerodia.logger.deprecate("using the 'class_name' locator to locate multiple "
                                         "classes with a str value (i.e. {!r})".format(value),
                                         "list (e.g. {!r})".format(value.split()),
                                         ids=['class_array'])
            return self._build_class_match(value)
        elif key == 'label' and self.should_use_label_element:
            # we assume 'label' means a corresponding label element, not the attribute
            text = 'normalize-space()={}'.format(XpathSupport.escape(value))
            return '(@id=//label[{0}]/@for or parent::label[{0}])'.format(text)
        else:
            return '{}={}'.format(self.lhs_for(building, key), XpathSupport.escape(value))

    # TODO: Get rid of building
    @staticmethod
    def lhs_for(building, key):
        if key == 'text':
            return 'normalize-space()'
        elif key == 'href':
            # TODO: change this behaviour?
            return 'normalize-space(@href)'
        elif key == 'type':
            # type attributes can be upper case - downcase them
            # https://github.com/watir/watir/issues/72
            return XpathSupport.lower('@type')
        else:
            return '@{}'.format(key.replace('_', '-'))

    # private

    def _process_adjacent(self, adjacent):
        xpath = './'
        if adjacent == 'ancestor':
            xpath += 'ancestor::'
        elif adjacent == 'preceding':
            xpath += 'preceding-sibling::'
        elif adjacent == 'following':
            xpath += 'following-sibling::'
        elif adjacent == 'child':
            xpath += 'child::'
        return xpath

    def _build_class_match(self, value):
        if re.match('!', value):
            escaped = XpathSupport.escape(" {} ".format(value[1:]))
            return "not(contains(concat(' ', @class, ' '), {}))".format(escaped)
        else:
            escaped = XpathSupport.escape(" {} ".format(value))
            return "contains(concat(' ', @class, ' '), {})".format(escaped)

    def _attribute_presence(self, attribute):
        return self.lhs_for(None, attribute)

    def _attribute_absence(self, attribute):
        return "not({})".format(self.lhs_for(None, attribute))



import nerodia
from .timer import Timer


class Wait(object):
    INTERVAL = 0.1
    timer = Timer()  # Access timer implementation in use

    @classmethod
    def until(cls, method=None, timeout=None, message=None, interval=None, object=None):
        """
        Waits until the method evaluates to True or times out
        :param method: method to run, typically lambda
        :param object: object to evaluate method with
        :param timeout: time to wait
        :type timeout: float
        :param message:  message to raise if timeout is exceeded
        :type message: str
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        Wait.until(lambda: browser.text_field(name='abrakadbra').present)
        """
        timeout = timeout or nerodia.default_timeout
        result = cls._run_with_timer(timeout, interval, method, object, until=True)
        if result:
            return result
        raise TimeoutError(cls._message_for(timeout, object, message))

    @classmethod
    def until_not(cls, method=None, timeout=None, message=None, interval=None, object=None):
        """
        Waits while the method evaluates to True or times out
        :param method: method to run, typically lambda
        :param object: object to evaluate method with
        :param timeout: time to wait
        :type timeout: float
        :param message:  message to raise if timeout is exceeded
        :type message: str
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        Wait.until_not(lambda: browser.text_field(name='abrakadbra').present)
        """
        timeout = timeout or nerodia.default_timeout
        result = cls._run_with_timer(timeout, interval, method, object, until=False)
        if result:
            return result
        raise TimeoutError(cls._message_for(timeout, object, message))

    whilst = until_not

    @classmethod
    def _message_for(cls, timeout, object, message):
        if callable(message):
            message = message(object)
        err = 'timed out after {} seconds'.format(timeout)
        if message:
            err += ', {}'.format(message)
        return err

    @classmethod
    def _run_with_timer(cls, timeout, interval, method, object, until=True):
        if timeout == 0:
            return method(object) if object else method()
        else:
            interval = interval or cls.INTERVAL
            result = cls.timer.wait(timeout, method, interval=interval,
                                    object=object, expected=until)
            return result


class Waitable(object):
    def wait_until(self, method=None, timeout=None, message=None, interval=None, object=None):
        """
        Waits until the condition is True
        :param method: method to run, typically lambda
        :param object: object to evaluate method with
        :param timeout: time to wait
        :type timeout: int
        :param message:  message to raise if timeout is exceeded
        :type message: str
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        browser.text_field(name='new_user_first_name').wait_until(lambda x: x.present).click
        """
        if not message:
            def msg(obj):
                return 'waiting for true condition on {}'.format(obj)
            message = msg
        if object is None:
            object = self
        Wait.until(method=method, timeout=timeout, message=message, interval=interval,
                   object=object)
        return self

    def wait_until_not(self, method=None, timeout=None, message=None, interval=None, object=None):
        """
        Waits while the condition is True
        :param method: method to run, typically lambda
        :param object: object to evaluate method with
        :param timeout: time to wait
        :type timeout: int
        :param message: message to raise if timeout is exceeded
        :type message: str
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        browser.wait_while(lambda x: not x.exists, timeout=2)
        """
        if not message:
            def msg(obj):
                return 'waiting for false condition on {}'.format(obj)
            message = msg
        if object is None:
            object = self
        Wait.until_not(method=method, timeout=timeout, message=message, interval=interval,
                       object=object)
        return self

    def wait_until_present(self, timeout=None, interval=None):
        """
        Waits until the element is present
        :param timeout: time to wait
        :type timeout: int
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        browser.text_field(name='new_user_first_name').wait_until_present()
        """
        from nerodia.elements.element import Element
        if isinstance(self, Element):
            def func(e):
                e.reset()
                return e.present
            return self.wait_until(method=func, timeout=timeout, interval=interval)

        nerodia.logger.deprecate('{}#wait_until_present'.format(self.__class__.__name__),
                                 '{}#wait_until(method=lambda e: e.present)',
                                 ids=['wait_until_present'])
        return self.wait_until(method=lambda x: x.present, timeout=timeout, interval=interval)

    def wait_until_not_present(self, timeout=None, interval=None):
        """
        Waits while the element is present
        :param timeout: time to wait
        :type timeout: int
        :param interval: time to wait between each check
        :type interval: float
        :Example:

        browser.text_field(name='abrakadbra').wait_until_not_present
        """
        from nerodia.elements.element import Element
        if isinstance(self, Element):
            def func(e):
                e.reset()
                return e.present
            return self.wait_until_not(method=func, timeout=timeout, interval=interval)

        nerodia.logger.deprecate('{}#wait_until_not_present'.format(self.__class__.__name__),
                                 '{}#wait_until_not(method=lambda e: e.present)',
                                 ids=['wait_until_not_present'])
        return self.wait_until_not(method=lambda x: x.present, timeout=timeout, interval=interval)


class TimeoutError(Exception):
    pass



from nerodia import tag_to_class
from .elements.area import Area
from .elements.button import Button
from .elements.d_list import DList
from .elements.form import Form
from .elements.html_elements import HTMLElement, Audio, Base, Quote, Body, BR, Canvas, \
    TableCaption, TableCol, Data, DataList, Mod, Details, Dialog, Div, Embed, FieldSet, FrameSet, \
    Heading, Head, HR, Html, Keygen, Label, Legend, LI, Map, Menu, MenuItem, Meta, Meter, Object, \
    OptGroup, Output, Paragraph, Param, Pre, Progress, Script, Source, Span, Style, \
    Template, TableHeaderCell, Time, Title, Track, Video, AnchorCollection, \
    HTMLElementCollection, AreaCollection, AudioCollection, BaseCollection, QuoteCollection, \
    BodyCollection, BRCollection, ButtonCollection, CanvasCollection, TableCaptionCollection, \
    TableColCollection, DataCollection, DataListCollection, ModCollection, DetailsCollection, \
    DialogCollection, DivCollection, DListCollection, EmbedCollection, FieldSetCollection, \
    FormCollection, FrameSetCollection, HeadingCollection, HeadCollection, HRCollection, \
    HtmlCollection, IFrameCollection, ImageCollection, InputCollection, KeygenCollection, \
    LabelCollection, LegendCollection, LICollection, MapCollection, MenuCollection, \
    MenuItemCollection, MetaCollection, MeterCollection, ObjectCollection, OListCollection, \
    OptGroupCollection, OptionCollection, OutputCollection, ParagraphCollection, ParamCollection, \
    PreCollection, ProgressCollection, ScriptCollection, SelectCollection, SourceCollection, \
    SpanCollection, StyleCollection, TableDataCellCollection, TemplateCollection, TextAreaCollection, \
    TableCollection, TableHeaderCellCollection, TableSectionCollection, TimeCollection, TitleCollection, \
    TableRowCollection, TrackCollection, UListCollection, VideoCollection
from .elements.i_frame import IFrame
from .elements.image import Image
from .elements.input import Input
from .elements.link import Anchor
from .elements.list import OList, UList
from .elements.option import Option
from .elements.radio import Radio, RadioCollection
from .elements.radio_set import RadioSet
from .elements.select import Select
from .elements.table import Table
from .elements.table_data_cell import TableDataCell
from .elements.table_row import TableRow
from .elements.table_section import TableSection
from .elements.text_area import TextArea

tag_to_class['a'] = Anchor
tag_to_class['a_collection'] = AnchorCollection
tag_to_class['link'] = Anchor
tag_to_class['link_collection'] = AnchorCollection
tag_to_class['abbr'] = HTMLElement
tag_to_class['abbr_collection'] = HTMLElementCollection
tag_to_class['address'] = HTMLElement
tag_to_class['address_collection'] = HTMLElementCollection
tag_to_class['area'] = Area
tag_to_class['area_collection'] = AreaCollection
tag_to_class['article'] = HTMLElement
tag_to_class['article_collection'] = HTMLElementCollection
tag_to_class['aside'] = HTMLElement
tag_to_class['aside_collection'] = HTMLElementCollection
tag_to_class['audio'] = Audio
tag_to_class['audio_collection'] = AudioCollection
tag_to_class['b'] = HTMLElement
tag_to_class['b_collection'] = HTMLElementCollection
tag_to_class['base'] = Base
tag_to_class['base_collection'] = BaseCollection
tag_to_class['bdi'] = HTMLElement
tag_to_class['bdi_collection'] = HTMLElementCollection
tag_to_class['bdo'] = HTMLElement
tag_to_class['bdo_collection'] = HTMLElementCollection
tag_to_class['blockquote'] = Quote
tag_to_class['blockquote_collection'] = QuoteCollection
tag_to_class['body'] = Body
tag_to_class['body_collection'] = BodyCollection
tag_to_class['br'] = BR
tag_to_class['br_collection'] = BRCollection
tag_to_class['button'] = Button
tag_to_class['button_collection'] = ButtonCollection
tag_to_class['canvas'] = Canvas
tag_to_class['canvas_collection'] = CanvasCollection
tag_to_class['caption'] = TableCaption
tag_to_class['caption_collection'] = TableCaptionCollection
tag_to_class['cite'] = HTMLElement
tag_to_class['cite_collection'] = HTMLElementCollection
tag_to_class['code'] = HTMLElement
tag_to_class['code_collection'] = HTMLElementCollection
tag_to_class['col'] = TableCol
tag_to_class['col_collection'] = TableColCollection
tag_to_class['colgroup'] = TableCol
tag_to_class['colgroup_collection'] = TableColCollection
tag_to_class['data'] = Data
tag_to_class['data_collection'] = DataCollection
tag_to_class['datalist'] = DataList
tag_to_class['datalist_collection'] = DataListCollection
tag_to_class['dd'] = HTMLElement
tag_to_class['dd_collection'] = HTMLElementCollection
tag_to_class['del'] = Mod
tag_to_class['del_collection'] = ModCollection
tag_to_class['delete'] = Mod
tag_to_class['delete_collection'] = ModCollection
tag_to_class['details'] = Details
tag_to_class['details_collection'] = DetailsCollection
tag_to_class['dfn'] = HTMLElement
tag_to_class['dfn_collection'] = HTMLElementCollection
tag_to_class['dialog'] = Dialog
tag_to_class['dialog_collection'] = DialogCollection
tag_to_class['div'] = Div
tag_to_class['div_collection'] = DivCollection
tag_to_class['dl'] = DList
tag_to_class['dl_collection'] = DListCollection
tag_to_class['dt'] = HTMLElement
tag_to_class['dt_collection'] = HTMLElementCollection
tag_to_class['em'] = HTMLElement
tag_to_class['em_collection'] = HTMLElementCollection
tag_to_class['embed'] = Embed
tag_to_class['embed_collection'] = EmbedCollection
tag_to_class['fieldset'] = FieldSet
tag_to_class['fieldset_collection'] = FieldSetCollection
tag_to_class['figcaption'] = HTMLElement
tag_to_class['figcaption_collection'] = HTMLElementCollection
tag_to_class['figure'] = HTMLElement
tag_to_class['figure_collection'] = HTMLElementCollection
tag_to_class['footer'] = HTMLElement
tag_to_class['footer_collection'] = HTMLElementCollection
tag_to_class['form'] = Form
tag_to_class['form_collection'] = FormCollection
tag_to_class['frameset'] = FrameSet
tag_to_class['frameset_collection'] = FrameSetCollection
tag_to_class['h1'] = Heading
tag_to_class['h1_collection'] = HeadingCollection
tag_to_class['h2'] = Heading
tag_to_class['h2_collection'] = HeadingCollection
tag_to_class['h3'] = Heading
tag_to_class['h3_collection'] = HeadingCollection
tag_to_class['h4'] = Heading
tag_to_class['h4_collection'] = HeadingCollection
tag_to_class['h5'] = Heading
tag_to_class['h5_collection'] = HeadingCollection
tag_to_class['h6'] = Heading
tag_to_class['h6_collection'] = HeadingCollection
tag_to_class['head'] = Head
tag_to_class['head_collection'] = HeadCollection
tag_to_class['header'] = HTMLElement
tag_to_class['header_collection'] = HTMLElementCollection
tag_to_class['hgroup'] = HTMLElement
tag_to_class['hgroup_collection'] = HTMLElementCollection
tag_to_class['hr'] = HR
tag_to_class['hr_collection'] = HRCollection
tag_to_class['html'] = Html
tag_to_class['html_collection'] = HtmlCollection
tag_to_class['i'] = HTMLElement
tag_to_class['i_collection'] = HTMLElementCollection
tag_to_class['ital'] = HTMLElement
tag_to_class['ital_collection'] = HTMLElementCollection
tag_to_class['iframe'] = IFrame
tag_to_class['iframe_collection'] = IFrameCollection
tag_to_class['img'] = Image
tag_to_class['img_collection'] = ImageCollection
tag_to_class['input'] = Input
tag_to_class['input_collection'] = InputCollection
tag_to_class['ins'] = Mod
tag_to_class['ins_collection'] = ModCollection
tag_to_class['kbd'] = HTMLElement
tag_to_class['kbd_collection'] = HTMLElementCollection
tag_to_class['keygen'] = Keygen
tag_to_class['keygen_collection'] = KeygenCollection
tag_to_class['label'] = Label
tag_to_class['label_collection'] = LabelCollection
tag_to_class['legend'] = Legend
tag_to_class['legend_collection'] = LegendCollection
tag_to_class['li'] = LI
tag_to_class['li_collection'] = LICollection
tag_to_class['main'] = HTMLElement
tag_to_class['main_collection'] = HTMLElementCollection
tag_to_class['map'] = Map
tag_to_class['map_collection'] = MapCollection
tag_to_class['mark'] = HTMLElement
tag_to_class['mark_collection'] = HTMLElementCollection
tag_to_class['menu'] = Menu
tag_to_class['menu_collection'] = MenuCollection
tag_to_class['menuitem'] = MenuItem
tag_to_class['menuitem_collection'] = MenuItemCollection
tag_to_class['meta'] = Meta
tag_to_class['meta_collection'] = MetaCollection
tag_to_class['meter'] = Meter
tag_to_class['meter_collection'] = MeterCollection
tag_to_class['nav'] = HTMLElement
tag_to_class['nav_collection'] = HTMLElementCollection
tag_to_class['noscript'] = HTMLElement
tag_to_class['noscript_collection'] = HTMLElementCollection
tag_to_class['object'] = Object
tag_to_class['object_collection'] = ObjectCollection
tag_to_class['ol'] = OList
tag_to_class['ol_collection'] = OListCollection
tag_to_class['optgroup'] = OptGroup
tag_to_class['optgroup_collection'] = OptGroupCollection
tag_to_class['option'] = Option
tag_to_class['option_collection'] = OptionCollection
tag_to_class['output'] = Output
tag_to_class['output_collection'] = OutputCollection
tag_to_class['p'] = Paragraph
tag_to_class['p_collection'] = ParagraphCollection
tag_to_class['param'] = Param
tag_to_class['param_collection'] = ParamCollection
tag_to_class['pre'] = Pre
tag_to_class['pre_collection'] = PreCollection
tag_to_class['progress'] = Progress
tag_to_class['progress_collection'] = ProgressCollection
tag_to_class['q'] = Quote
tag_to_class['q_collection'] = QuoteCollection
tag_to_class['radio'] = Radio
tag_to_class['radio_collection'] = RadioCollection
tag_to_class['radio_set'] = RadioSet
tag_to_class['rp'] = HTMLElement
tag_to_class['rp_collection'] = HTMLElementCollection
tag_to_class['rt'] = HTMLElement
tag_to_class['rt_collection'] = HTMLElementCollection
tag_to_class['ruby'] = HTMLElement
tag_to_class['ruby_collection'] = HTMLElementCollection
tag_to_class['s'] = HTMLElement
tag_to_class['s_collection'] = HTMLElementCollection
tag_to_class['samp'] = HTMLElement
tag_to_class['samp_collection'] = HTMLElementCollection
tag_to_class['script'] = Script
tag_to_class['script_collection'] = ScriptCollection
tag_to_class['section'] = HTMLElement
tag_to_class['section_collection'] = HTMLElementCollection
tag_to_class['select'] = Select
tag_to_class['select_collection'] = SelectCollection
tag_to_class['select_list'] = Select
tag_to_class['select_list_collection'] = SelectCollection
tag_to_class['small'] = HTMLElement
tag_to_class['small_collection'] = HTMLElementCollection
tag_to_class['source'] = Source
tag_to_class['source_collection'] = SourceCollection
tag_to_class['span'] = Span
tag_to_class['span_collection'] = SpanCollection
tag_to_class['strong'] = HTMLElement
tag_to_class['strong_collection'] = HTMLElementCollection
tag_to_class['style'] = Style
tag_to_class['style_collection'] = StyleCollection
tag_to_class['sub'] = HTMLElement
tag_to_class['sub_collection'] = HTMLElementCollection
tag_to_class['summary'] = HTMLElement
tag_to_class['summary_collection'] = HTMLElementCollection
tag_to_class['sup'] = HTMLElement
tag_to_class['sup_collection'] = HTMLElementCollection
tag_to_class['table'] = Table
tag_to_class['table_collection'] = TableCollection
tag_to_class['tbody'] = TableSection
tag_to_class['tbody_collection'] = TableSectionCollection
tag_to_class['td'] = TableDataCell
tag_to_class['td_collection'] = TableDataCellCollection
tag_to_class['template'] = Template
tag_to_class['template_collection'] = TemplateCollection
tag_to_class['textarea'] = TextArea
tag_to_class['textarea_collection'] = TextAreaCollection
tag_to_class['tfoot'] = TableSection
tag_to_class['tfoot_collection'] = TableSectionCollection
tag_to_class['th'] = TableHeaderCell
tag_to_class['th_collection'] = TableHeaderCellCollection
tag_to_class['thead'] = TableSection
tag_to_class['thead_collection'] = TableSectionCollection
tag_to_class['time'] = Time
tag_to_class['time_collection'] = TimeCollection
tag_to_class['title'] = Title
tag_to_class['title_collection'] = TitleCollection
tag_to_class['tr'] = TableRow
tag_to_class['tr_collection'] = TableRowCollection
tag_to_class['track'] = Track
tag_to_class['track_collection'] = TrackCollection
tag_to_class['u'] = HTMLElement
tag_to_class['u_collection'] = HTMLElementCollection
tag_to_class['ul'] = UList
tag_to_class['ul_collection'] = UListCollection
tag_to_class['var'] = HTMLElement
tag_to_class['var_collection'] = HTMLElementCollection
tag_to_class['video'] = Video
tag_to_class['video_collection'] = VideoCollection
tag_to_class['wbr'] = HTMLElement
tag_to_class['wbr_collection'] = HTMLElementCollection



import six

from .table_cell import TableCell
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class TableDataCell(TableCell):
    _attr_abbr = (str, 'abbr')
    pass



from . import html_attributes, svg_attributes


class MetaHTMLElement(type):
    def __new__(cls, name, parents, dct):
        final_dict = create_attributes(name, parents, dct, html_attributes)

        inst = super(MetaHTMLElement, cls).__new__(cls, name, parents, final_dict)

        return inst


class MetaSVGElement(type):
    def __new__(cls, name, parents, dct):
        final_dict = create_attributes(name, parents, dct, svg_attributes)

        inst = super(MetaSVGElement, cls).__new__(cls, name, parents, final_dict)

        return inst


def create_attributes(name, parents, dct, generated):
    final_dict = {}

    attrs = dct.get('ATTRIBUTES', [])
    for key, value in dct.items():
        if key.startswith('_attr'):
            attr_name = key.split('_attr_')[-1]
            typ, val = value
            final_dict[attr_name] = property(fget=make_attr(typ=typ, val=val))
            attrs.append(attr_name)
        elif key.startswith('_aliases'):
            continue
        else:
            final_dict[key] = value
    auto_gen = getattr(generated, name.lower(), [])
    for each in auto_gen:
        typ, key, val = each
        final_dict[key] = property(fget=make_attr(typ, val))
        attrs.append(key)

    for parent in parents:
        attrs.extend(getattr(parent, 'ATTRIBUTES', []))

    final_dict['ATTRIBUTES'] = list(set(attrs))
    return final_dict


def make_attr(typ, val):
    if typ == bool:
        def attr_bool(self):
            return self.attribute_value(val) == 'true'

        return attr_bool

    elif typ == int:
        def attr_int(self):
            value = self.attribute_value(val)
            return value and int(value)

        return attr_int

    elif typ == float:
        def attr_float(self):
            value = self.attribute_value(val)
            return value and float(value)

        return attr_float

    else:
        def attr_str(self):
            return str(self.attribute_value(val) or '')

        return attr_str



import six

from .html_elements import TableCellCollection
from .table_cell import TableCell
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Cell(TableCell):
    pass


@six.add_metaclass(MetaHTMLElement)
class CellCollection(TableCellCollection):
    @property
    def _elements(self):
        # we do this craziness since the xpath used will find direct child rows
        # before any rows inside thead/tbody/tfoot...
        elements = super(CellCollection, self)._elements
        return sorted(elements, key=lambda e: int(e.get_attribute('cellIndex')))



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Input(HTMLElement):
    def label(self):
        """
        Returns label element associated with Input element.

        :rtype: nerodia.elements.html_elements.Label
        """
        return self.parent(tag_name='form').label(**{'for': self.id})



import re
from .elements.cell import Cell, CellCollection


class CellContainer(object):
    def cell(self, *args, **kwargs):
        return Cell(self, dict(self._extract_selector(*args, **kwargs),
                               tag_name=re.compile('^(th|td)$')))

    def cells(self, *args, **kwargs):
        return CellCollection(self, dict(self._extract_selector(*args, **kwargs),
                                         tag_name=re.compile('^(th|td)$')))



import re
from time import sleep

import nerodia


class JSExecution(object):
    def execute_script(self, script, *args):
        """
        Delegates script execution to Browser or IFrame

        :param script: script to execute
        """
        return self.query_scope.execute_script(script, *args)

    def fire_event(self, event_name):
        """
        Simulates JavaScript events on element.
        Note that you may omit "on" from event name.

        :param event_name: event to fire

        :Example:

        browser.button(name: 'new_user_button').fire_event('click')
        browser.button(name: 'new_user_button').fire_event('mousemove')
        browser.button(name: 'new_user_button').fire_event('onmouseover')
        """
        event_name = re.sub(r'^on', '', event_name).lower()
        return self._element_call(lambda: self._execute_js('fireEvent', self, event_name))

    def flash(self, preset='default', color='red', flashes=10, delay=0.05):
        """
        Flashes (change background color to a new color and back a few times) element

        :param preset: choice from preset values
        :type param: str:
        :param color: what color or colors to flash with
        :type color: str or list
        :param flashes: number of times element should be flashed
        :type flashes: int
        :param delay: how long to wait between flashes
        :type delay: int or float

        :returns: nerodia.elements.element.Element

        :Example:

        browser.li(id='non_link_1').flash()
        browser.li(id='non_link_1').flash(color='green', flashes=3, delay=0.05)
        browser.li(id='non_link_1').flash(color='yellow')
        browser.li(id='non_link_1').flash(flashes=4)
        browser.li(id='non_link_1').flash(delay=0.1)
        """
        presets = {
            'fast': {'delay': 0.04},
            'slow': {'delay': 0.2},
            'long': {'flashes': 5, 'delay': 0.5},
            'rainbow': {'flashes': 5, 'color': ['red', 'orange', 'yellow', 'green', 'blue',
                                                'indigo', 'violet']}
        }
        # TODO: remove this sooner than later
        if preset in presets:
            return self.flash(**presets[preset])
        elif isinstance(color, int):
            nerodia.logger.deprecate('Using color as first argument to #flash',
                                     'Specify with keyword or add preset')
            if isinstance(flashes, float):
                delay = flashes
            flashes = color
            color = preset

        background_color = self.style('background-color')
        original_color = background_color
        if not background_color:
            background_color = 'white'

        if not isinstance(color, list):
            colors = [color]
        else:
            colors = color[:]
        colors.append(background_color)

        for next_color in colors * flashes:
            self._element_call(lambda: self._execute_js('backgroundColor', self.el, next_color))
            sleep(delay)

        self._element_call(lambda: self._execute_js('backgroundColor', self.el, original_color))

        return self

    def focus(self):
        """
        Focuses the element
        Note that Firefox queues focus events until the window actually has focus
        """
        return self._element_call(lambda: self._execute_js('focus', self.el))

    @property
    def inner_html(self):
        """
        Returns inner HTML code of element

        :rtype: str

        :Example:

        browser.div(id='shown').inner_html
        #=> '<div id="hidden" style="display: none;">Not shown</div><div>Not hidden</div>'
        """
        return self._element_call(lambda: self._execute_js('getInnerHtml', self.el)).strip()

    @property
    def inner_text(self):
        """
        Returns inner Text code of element

        :rtype: str:

        :Example:

        browser.div(id: 'shown').inner_text
        #=> "Not hidden"
        """
        return self._element_call(lambda: self._execute_js('getInnerText', self)).strip()

    @property
    def outer_html(self):
        """
        Returns outer (inner + element itself) HTML code of element

        :rtype: str

        :Example:

        browser.div(id='shown').outer_html
        #=> '<div id="shown"><div id="hidden" style="display: none;">Not shown</div><div>Not hidden</div></div>'
        """
        return self._element_call(lambda: self._execute_js('getOuterHtml', self.el)).strip()

    html = outer_html

    @property
    def text_content(self):
        """
        Returns text content of element

        :rtype: str

        :Example:

        browser.div(id='shown').text_content
        #=> 'Not shownNot hidden'
        """
        return self._element_call(lambda: self._execute_js('getTextContent', self.el)).strip()

    def select_text(self, string):
        """
        Selects text on page (as if dragging clicked mouse across provided text)

        :param string: string to select

        :Example:

        browser.li(id='non_link_1').select_text('Non-link')
        """
        return self._element_call(lambda: self._execute_js('selectText', self.el, string))



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement
from ..row_container import RowContainer


@six.add_metaclass(MetaHTMLElement)
class TableSection(RowContainer, HTMLElement):
    def __getitem__(self, idx):
        """
        Returns row of this table with given index

        :param idx: row index
        :rtype: nerodia.elements.row.Row
        """
        return self.row(index=idx)



class RowContainer(object):
    def row(self, *args, **kwargs):
        from .elements.row import Row
        return Row(self, self._extract_selector(tag_name='tr', *args, **kwargs))

    def rows(self, *args, **kwargs):
        from .elements.row import RowCollection
        return RowCollection(self, self._extract_selector(tag_name='tr', *args, **kwargs))

    @property
    def strings(self):
        """
        A table as a 2D array of strings with the text of each cell
        :rtype: list[list[str]]
        """
        self.wait_for_exists()

        rows = []
        for row in self.rows():
            rows.append([cell.text for cell in row])
        return rows



from importlib import import_module


class ClassHelpers(object):
    # private

    @property
    def _locator_class(self):
        from .element.locator import Locator
        return getattr(self._import_module, 'Locator', Locator)

    @property
    def _element_validator_class(self):
        from .element.validator import Validator
        return getattr(self._import_module, 'Validator', Validator)

    @property
    def _selector_builder_class(self):
        from .element.selector_builder import SelectorBuilder
        return getattr(self._import_module, 'SelectorBuilder', SelectorBuilder)

    @property
    def _import_module(self):
        from ..module_mapping import map_module
        modules = [self.browser.locator_namespace.__name__, map_module(self._element_class_name)]
        try:
            return import_module('{}.{}'.format(*modules))
        except ImportError:
            return import_module('{}.element'.format(*modules[:1]))

    @property
    def _element_class_name(self):
        return self._element_class.__name__

    def _build_locator(self):
        self.query_scope._ensure_context()

        element_validator = self._element_validator_class()
        selector_builder = self._selector_builder_class(self.query_scope, self.selector.copy(),
                                                        self._element_class.ATTRIBUTES)
        return self._locator_class(self.query_scope, self.selector.copy(), selector_builder,
                                   element_validator)



import re

MODULE_MAPPING = {'frame': 'i_frame',
                  'anchor': 'link',
                  'o_list': 'list',
                  'u_list': 'list'}


def map_module(name):
    element_module = re.sub(r'([A-Z]{1})', r'_\1', name)[1:].lower()
    if element_module in list(MODULE_MAPPING):  # special cases
        return MODULE_MAPPING.get(element_module)
    else:
        return element_module



import six

from .html_elements import TableRowCollection
from .table_row import TableRow
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Row(TableRow):
    pass


@six.add_metaclass(MetaHTMLElement)
class RowCollection(TableRowCollection):
    def __init__(self, *args, **kwargs):
        self.list = None
        super(RowCollection, self).__init__(*args, **kwargs)

    def __iter__(self):
        """
        Yields each row in collection

        :rtype: iter

        :Example:

        rows = browser.rows(class='kls')
        for row in rows:
             print(row.text)
        """
        # we do this craziness since the xpath used will find direct child rows
        # before any rows inside thead/tbody/tfoot...
        if not self.list:
            self._get_list()
        for item in self.list:
            yield item

    @property
    def to_list(self):
        # we do this craziness since the xpath used will find direct child rows
        # before any rows inside thead/tbody/tfoot...
        import nerodia
        nerodia.logger.deprecate('RowCollection.to_list', 'list(RowCollection)')
        return list(self)

    # private

    def _get_list(self):
        if not self.list:
            elements = list(super(RowCollection, self).__iter__())
            self.list = sorted(elements, key=lambda e: int(e.attribute_value('rowIndex')))



from inspect import stack
from re import search, sub

from selenium.common.exceptions import InvalidElementStateException, \
    StaleElementReferenceException, \
    ElementNotVisibleException, ElementNotInteractableException, NoSuchWindowException
from selenium.webdriver.common.action_chains import ActionChains

import nerodia
from nerodia.browser import Browser
from ..adjacent import Adjacent
from ..container import Container
from ..exception import Error, ObjectDisabledException, ObjectReadOnlyException, \
    UnknownFrameException, UnknownObjectException, NoMatchingWindowFoundException
from ..js_execution import JSExecution
from ..js_snippet import JSSnippet
from ..locators.class_helpers import ClassHelpers
from ..locators.element.selector_builder import SelectorBuilder
from ..wait.wait import TimeoutError, Wait, Waitable
from ..window import Dimension, Point


class Element(ClassHelpers, JSExecution, Container, JSSnippet, Waitable, Adjacent):
    ATTRIBUTES = []

    def __init__(self, query_scope, selector):
        self.query_scope = query_scope
        if not isinstance(selector, dict):
            raise TypeError('invalid argument: {!r}'.format(selector))

        self.el = selector.pop('element', None)
        self.selector = selector
        self.keyword = None
        self.locator = None

    @property
    def exists(self):
        """
        Returns True if element exists, False otherwise
        :rtype: bool
        """
        try:
            if self.el and self.stale:
                return False
            self.assert_exists()
            return True
        except (UnknownObjectException, UnknownFrameException):
            return False

    exist = exists

    def __repr__(self):
        string = '#<{}: '.format(self.__class__.__name__)
        if self.keyword:
            string += 'keyword: {} '.format(self.keyword)
        string += 'located: {}; '.format(self.el is not None)
        if not self.selector:
            string += '{element: (selenium element)}'
        else:
            string += self.selector_string
        string += '>'
        return string

    def __eq__(self, other):
        """
        Returns True if two elements are equal
        :param other: other element to compare
        :rtype: bool
        """
        return isinstance(other, self.__class__) and self.wd == other.wd

    eql = __eq__

    def __hash__(self):
        return self.el.__hash__() if self.el else super(Element, self).__hash__()

    @property
    def text(self):
        """
        Returns the text of the element
        :rtype: str
        """
        return self._element_call(lambda: self.el.text)

    @property
    def tag_name(self):
        """
        Returns the tag name of the element
        :rtype: str
        """
        return self._element_call(lambda: self.el.tag_name).lower()

    def click(self, *modifiers):
        """
        Clicks the element, optionally while pressing the given modifier keys.
        Note that support for holding a modifier key is currently experimental, and may not work
        at all.

        :param modifiers: modifier keys to press while clicking
        :Example: Click an element

        browser.element(name='new_user_button').click()

        :Example: Click an element with shift key pressed
        from selenium.webdriver.common.keys import Keys
        browser.element(name='new_user_button').click(Keys.SHIFT)

        :Example: Click an element with several modifier keys pressed
        from selenium.webdriver.common.keys import Keys
        browser.element(name='new_user_button').click(Keys.SHIFT, Keys.CONTROL)
        """

        def method():
            if modifiers:
                action = ActionChains(self.driver)
                for mod in modifiers:
                    action.key_down(mod)
                action.click(self.el)
                for mod in modifiers:
                    action.key_up(mod)
                action.perform()
            else:
                self.el.click()

        self._element_call(method, self.wait_for_enabled)
        self.browser.after_hooks.run()

    def js_click(self):
        """
        Simulates JavaScript click event on element.

        :Example: Click an element
        browser.element(name='new_user_button').js_click()
        """
        self.fire_event('click')
        self.browser.after_hooks.run()

    def double_click(self):
        """
        Double clicks the element.
        Note that browser support may vary.

        :Example: Double-click an element

        browser.element(name='new_user_button').double_click()
        """
        self._element_call(lambda: ActionChains(self.driver).double_click(self.el)
                           .perform(), self.wait_for_present)
        self.browser.after_hooks.run()

    def js_double_click(self):
        """
        Simulates JavaScript double click event on element.

        :Example: Click an element
        browser.element(name='new_user_button').js_double_click()
        """
        self.fire_event('dblclick')
        self.browser.after_hooks.run()

    def right_click(self):
        """
        Right clicks the element
        Note that browser support may vary

        :Example: Right click an element

        browser.element(name='new_user_button').right_click()
        """
        self._element_call(lambda: ActionChains(self.driver).context_click(self.el)
                           .perform(), self.wait_for_present)
        self.browser.after_hooks.run()

    def hover(self):
        """
        Moves the mouse to the middle of this element
        Note that browser support may vary

        :Example: Hover over an element

        browser.element(name='new_user_button').hover()
        """
        self._element_call(lambda: ActionChains(self.driver).move_to_element(self.el)
                           .perform(), self.wait_for_present)
        self.browser.after_hooks.run()

    def drag_and_drop_on(self, other):
        """
        Drag and drop this element on to another element instance
        Note that browser support may vary

        :param other: element to drop on

        :Example: Drag an element onto another

        a = browser.div(id='draggable')
        b = browser.div(id='droppable')
        a.drag_and_drop_on(b)
        """
        self._assert_is_element(other)

        value = self._element_call(lambda: ActionChains(self.driver)
                                   .drag_and_drop(self.el, other.wd).perform(),
                                   self.wait_for_present)
        self.browser.after_hooks.run()
        return value

    def drag_and_drop_by(self, xoffset, yoffset):
        """
        Drag and drop this element by the given offsets.
        Note that browser support may vary.

        :param xoffset: amount to move horizontally
        :param yoffset: amount to move vertically

        :Example: Drag an element onto another

        browser.div(id='draggable').drag_and_drop_by(100, -200)
        """
        self._element_call(lambda: ActionChains(self.driver).
                           drag_and_drop_by_offset(self.el, xoffset, yoffset).perform(),
                           self.wait_for_present)

    def select_text(self, string):
        """
        Selects text on page (as if dragging clicked mouse across provided text)

        :param string: string to select

        :Example:

        browser.legend().select_text('information')
        """
        self._element_call(lambda: self._execute_js('selectText', self.el, string))

    @property
    def value(self):
        """
        Returns value of the element
        :rtype: str
        """
        try:
            return self.attribute_value('value') or ''
        except InvalidElementStateException:
            return ''

    def attribute_value(self, attribute_name):
        """
        Returns given attribute value of the element

        :param attribute_name: attribute to retrieve
        :type attribute_name: str
        :rtype: str

        :Example:

        browser.a(id='link_2').attribute_value('title')  #=> 'link_title_2'
        """
        return self._element_call(lambda: self.el.get_attribute(attribute_name))

    get_attriubte = attribute_value

    attribute = attribute_value

    def send_keys(self, *args):
        """
        Sends sequence of keystrokes to the element
        :param args: keystrokes to send

        :Example:

        browser.text_field(name='new_user_first_name').send_keys('nerodia')
        """
        return self._element_call(lambda: self.el.send_keys(*args), self.wait_for_writable)

    @property
    def focused(self):
        """
        Returns True if the element is focused
        :rtype: bool
        """
        return self._element_call(lambda: self.el == self.driver.switch_to.active_element)

    def fire_event(self, event_name):
        """
        Simulates JavaScript events on element
        Note that you may omit 'on' from event name

        :param event_name: event to fire

        :Example:

        browser.button(name='new_user_button').fire_event('click')
        browser.button(name='new_user_button').fire_event('mousemove')
        browser.button(name='new_user_button').fire_event('onmouseover')
        """
        event_name = sub(r'^on', '', str(event_name)).lower()

        self._element_call(lambda: self._execute_js('fireEvent', self.el, event_name))

    def scroll_into_view(self):
        """
        Scroll until the element is in the view screen
        :rtype: Point

        :Example:

        browser.button(name='new_user_button').scroll_into_view()
        """
        return Point(**self._element_call(lambda: self.el.location_once_scrolled_into_view))

    @property
    def location(self):
        """
        Get the location of the element (x, y)
        :rtype: Point

        :Example:

        browser.button(name='new_user_button').location
        """
        return Point(**self._element_call(lambda: self.el.location))

    @property
    def size(self):
        """
        Get the size of the element (width, height)
        :rtype: Dimension

        :Example:

        browser.button(name='new_user_button').size
        """
        return Dimension(**self._element_call(lambda: self.el.size))

    @property
    def height(self):
        """
        Get the height of the element
        :rtype: int

        :Example:

        browser.button(name='new_user_button').height
        """
        return self.size.height

    @property
    def width(self):
        """
        Get the width of the element
        :rtype: int

        :Example:

        browser.button(name='new_user_button').width
        """
        return self.size.width

    @property
    def center(self):
        """
        Get the center coordinates of the element
        :rtype: Point

        :Example:

        browser.button(name='new_user_button').center
        """
        location = self.location
        size = self.size
        return Point(round(location.x + size.width / 2), round(location.y + size.height / 2))

    centre = center

    @property
    def driver(self):
        return self.query_scope.driver

    @property
    def wd(self):
        from .i_frame import FramedDriver
        if self.el is None:
            self.assert_exists()
        if isinstance(self.el, FramedDriver):
            return self.driver
        return self.el

    @property
    def visible(self):
        """
        Returns true if this element is visible on the page
        Raises exception if element does not exist

        :rtype: bool
        """
        try:
            nerodia.logger.warning('#visible behavior will be changing slightly, consider '
                                   'switching to #present (more details: '
                                   'http://watir.com/element-existentialism/',
                                   ids=['visible_element'])
            self.assert_exists()
            return self.el.is_displayed()
        except StaleElementReferenceException:
            nerodia.logger.deprecate('Checking `#visible is False` to determine a stale element',
                                     '`#stale is True`', ids=['stale_visible'])
            self.reset()
            raise self._unknown_exception

    @property
    def enabled(self):
        """
        Returns True if the element is present and enabled on the page

        :rtype: bool
        """
        return self._element_call(lambda: self.el.is_enabled(), self.assert_exists)

    @property
    def present(self):
        """
        Returns True if the element exists and is visible on the page
        Returns False if the element does not exist or exists but is not visible

        :rtype: bool
        """
        try:
            self.assert_exists()
            return self.el.is_displayed()
        except StaleElementReferenceException:
            nerodia.logger.deprecate('Checking `#present is False` to determine a stale element',
                                     '`#stale is True`', ids=['stale_present'])
            self.reset()
            return False
        except (UnknownObjectException, UnknownFrameException):
            return False

    def style(self, prop=None):
        """
        Returns given style property of this element

        :param prop: property to get
        :type prop: str
        :rtype: str

        :Example:

        browser.button(value='Delete').style           #=> "border: 4px solid red;"
        browser.button(value='Delete').style('border') #=> "4px solid rgb(255, 0, 0)"
        """
        if prop:
            return self._element_call(lambda: self.el.value_of_css_property(prop))
        else:
            return str(self.attribute_value('style')).strip()

    def to_subtype(self):
        """
        Cast this Element instance to a more specific subtype
        :Example:

        browser.element(xpath="//input[@type='submit']").to_subtype()  #=> #<Button>
        """
        tag = self.tag_name
        from .button import Button
        from .check_box import CheckBox
        from .file_field import FileField
        from .html_elements import HTMLElement
        from .radio import Radio
        from .text_field import TextField

        if tag == 'input':
            elem_type = self.attribute_value('type')
            if elem_type in Button.VALID_TYPES:
                klass = Button
            elif elem_type == 'checkbox':
                klass = CheckBox
            elif elem_type == 'radio':
                klass = Radio
            elif elem_type == 'file':
                klass = FileField
            else:
                klass = TextField
        else:
            klass = nerodia.element_class_for(tag) or HTMLElement

        return klass(self.query_scope, selector=dict(self.selector, element=self.wd))

    @property
    def browser(self):
        """
        Returns browser
        :rtype: nerodia.browser.Browser
        """
        return self.query_scope.browser

    @property
    def stale(self):
        """
        Returns True if a previously located element is no longer attached to the DOM
        :rtype: bool
        """
        if self.el is None:
            raise Error('Can not check staleness of unused element')
        if self.stale_in_context:
            self.query_scope._ensure_context()
            return self.stale_in_context
        else:
            return False

    @property
    def stale_in_context(self):
        try:
            self.el.is_enabled()  # any wire call will check for staleness
            return False
        except StaleElementReferenceException:
            return True

    def reset(self):
        self.el = None

    def wait_for_exists(self):
        if not nerodia.relaxed_locate:
            return self.assert_exists()
        if self.exists:  # Performance shortcut
            return None

        try:
            if not isinstance(self.query_scope, Browser):
                self.query_scope.wait_for_exists()
            self.wait_until(lambda e: e.exists)
        except TimeoutError:
            raise self._unknown_exception('timed out after {} seconds, waiting for {} to be '
                                          'located'.format(nerodia.default_timeout, self))

    def wait_for_present(self):
        pres = self.present
        if not nerodia.relaxed_locate or pres:
            return pres

        try:
            if not isinstance(self.query_scope, Browser):
                self.query_scope.wait_for_present()
            self.wait_until_present()
        except TimeoutError as e:
            raise self._unknown_exception('element located, but {}'.format(e))

    def wait_for_enabled(self):
        from .button import Button
        from .input import Input
        from .option import Option
        from .select import Select
        if not nerodia.relaxed_locate:
            return self._assert_enabled()

        self.wait_for_exists()
        if not any(isinstance(self, klass) for klass in [Input, Button, Select, Option]):
            return

        try:
            self.wait_until(lambda e: e.enabled)
        except TimeoutError:
            self._raise_disabled()

    def wait_for_writable(self):
        self.wait_for_exists()

        self.wait_for_enabled()
        if not nerodia.relaxed_locate:
            if hasattr(self, 'readonly') and self.readonly:
                self._raise_writable()

        try:
            self.wait_until(lambda e: not getattr(e, 'readonly', None) or not e.readonly)
        except TimeoutError:
            raise ObjectReadOnlyException('element present and enabled, but timed out after {} '
                                          'seconds, waiting for {} to not be '
                                          'readonly'.format(nerodia.default_timeout, self))

    def assert_exists(self):
        """
        Ensure that the element exists, making sure that it is not stale and located if necessary
        """
        if not self.el:
            self.locate()
        if self.el is None:
            raise UnknownObjectException('unable to locate element: {}'.format(self))

    def locate(self):
        self.locator = self._build_locator()

        self.el = self.locator.locate()
        return self.el

    @property
    def selector_string(self):
        from ..browser import Browser
        if isinstance(self.query_scope, Browser):
            return self.selector.__repr__()
        else:
            return '{} --> {}'.format(self.query_scope.selector_string, self.selector)

    # Private

    def _raise_writable(self):
        raise ObjectReadOnlyException('element present and enabled, but timed out after {} '
                                      'seconds, waiting for {} to not be '
                                      'readonly'.format(nerodia.default_timeout, self))

    def _raise_disabled(self):
        raise ObjectDisabledException('element present, but timed out after {} '
                                      'seconds, waiting for {} to be '
                                      'enabled'.format(nerodia.default_timeout, self))

    def _raise_present(self):
        raise UnknownObjectException('element located, but timed out after {} seconds, waiting '
                                     'for {} to be present'.format(nerodia.default_timeout, self))

    @property
    def _unknown_exception(self):
        return UnknownObjectException

    @property
    def _element_class(self):
        return self.__class__

    # Ensure the driver is in the desired browser context
    def _ensure_context(self):
        if not self.exists:
            self.locate()

    def _is_attribute(self, attribute_name):
        return self.attribute_value(attribute_name) is not None

    def _assert_enabled(self):
        if not self._element_call(lambda: self.el.is_enabled()):
            raise ObjectDisabledException('object is disabled {}'.format(self))

    @classmethod
    def _assert_is_element(cls, obj):
        if not isinstance(obj, Element):
            raise TypeError('execpted nerodia.Element, '
                            'got {}:{}'.format(obj, obj.__class__.__name__))

    def _element_call(self, method, precondition=None):
        caller = stack()[1][3]
        already_locked = Wait.timer.locked
        if not already_locked:
            from ..wait.timer import Timer
            Wait.timer = Timer(timeout=nerodia.default_timeout)
        try:
            return self._element_call_check(precondition, method, caller)
        finally:
            nerodia.logger.info('<- `Completed {}#{}`'.format(self, caller))
            if not already_locked:
                Wait.timer.reset()

    def _check_condition(self, condition):
        nerodia.logger.info('<- `Verifying precondition {}#{}`'.format(self, condition))
        try:
            if not condition:
                self.assert_exists()
            else:
                condition()
            nerodia.logger.info('<- `Verified precondition '
                                '{}#{!r}`'.format(self, condition or 'assert_exists'))
        except self._unknown_exception:
            if condition is None:
                nerodia.logger.info('<- `Unable to satisfy precondition '
                                    '{}#{}`'.format(self, condition))
                self._check_condition(self.wait_for_exists)
            else:
                raise

    def _element_call_check(self, precondition, method, caller):
        nerodia.logger.info('-> `Executing {}#{}`'.format(self, caller))
        while True:
            try:
                self._check_condition(precondition)
                return method()
            except self._unknown_exception as e:
                if precondition is None:
                    self._element_call(method, self.wait_for_exists)
                msg = str(e)
                if self.query_scope._ensure_context() and len(self.query_scope.iframes()) > 0:
                    msg += '; Maybe look in an iframe?'
                custom_attributes = []
                if self.locator:
                    custom_attributes = self.locator.selector_builder.custom_attributes
                if custom_attributes:
                    msg += '; Nerodia treated {!r} as a non-HTML compliant attribute, ' \
                           'ensure that was intended'.format(custom_attributes)
                raise self._unknown_exception(msg)
            except StaleElementReferenceException:
                self.query_scope._ensure_context()
                self.reset()
                self._check_condition(precondition)
                return method()
            except (ElementNotVisibleException, ElementNotInteractableException):
                if (Wait.timer.remaining_time <= 0) or \
                        (precondition not in [self.wait_for_present, self.wait_for_enabled,
                                              self.wait_for_writable]):
                    self._raise_present()
                continue
            except InvalidElementStateException:
                if (Wait.timer.remaining_time <= 0) or \
                        (precondition in [self.wait_for_writable, self.wait_for_enabled,
                                          self.wait_for_writable]):
                    self._raise_disabled()
                continue
            except NoSuchWindowException:
                raise NoMatchingWindowFoundException('browser window was closed')

    def __getattribute__(self, name):
        if search(SelectorBuilder.WILDCARD_ATTRIBUTE, name):
            return self.attribute_value(name.replace('_', '-'))
        else:
            return object.__getattribute__(self, name)



import six

from ..elements.html_elements import LICollection, HTMLElement
from ..meta_elements import MetaHTMLElement


class List(object):
    def __iter__(self):
        """
        Yields each LI associated with this list

        :rtype: iter

        :Example:

        divs = browser.divs(class='kls')
        for li in browser.ol():
             print(li.text)
        """
        for e in self.list_items:
            yield e

    def __getitem__(self, idx):
        """
        Returns item from this list at the given index

        :param idx: index of wanted element, 0-indexed
        :type idx: int
        :return: instance of Element subclass
        :rtype: nerodia.elements.element.Element
        """
        return self.list_items[idx]

    @property
    def list_items(self):
        return LICollection(self, {'adjacent': 'child', 'tag_name': 'li'})


@six.add_metaclass(MetaHTMLElement)
class OList(HTMLElement, List):
    pass


@six.add_metaclass(MetaHTMLElement)
class UList(HTMLElement, List):
    pass



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Font(HTMLElement):

    @property
    def size(self):
        """
        Returns the size of the font
        :rtype: int

        :Example:

        browser.font().size    #=> 12
        """
        return self.attribute_value('size')



import six

from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Image(HTMLElement):
    @property
    def loaded(self):
        """
        Returns True if the image is loaded
        :rtype: bool
        """
        if not self.complete:
            return False

        return self._element_call(lambda: self._execute_js('isImageLoaded', self.el))



from errno import ENOENT
from os import path

import six

from .html_elements import InputCollection
from .input import Input
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class FileField(Input):
    def set(self, filepath):
        """
        Set the file field to the given path

        :param filepath: path to the file
        :raises: ENOENT
        """
        if not path.exists(filepath):
            raise OSError(ENOENT, '{!r} does not exist.'.format(filepath))
        self.value = filepath

    @property
    def value(self):
        """
        Gets teh value of the file field
        :rtype: str
        """
        return self.attribute_value('value')

    @value.setter
    def value(self, filepath):
        """
        Set the file field to the given path
        :param filepath: path to the file
        """
        self._element_call(lambda: self.el.send_keys(filepath))


@six.add_metaclass(MetaHTMLElement)
class FileFieldCollection(InputCollection):
    pass



import re

import six

import nerodia
from nerodia.exception import Error, NoValueFoundException, UnknownObjectException, \
    ObjectDisabledException
from nerodia.wait.wait import TimeoutError
from .html_elements import HTMLElement
from ..meta_elements import MetaHTMLElement
from ..wait.wait import Wait

try:
    from re import Pattern
except ImportError:
    from re import _pattern_type as Pattern


@six.add_metaclass(MetaHTMLElement)
class Select(HTMLElement):
    def clear(self):
        """ Clears all selected options """
        if not self.multiple:
            raise Error('you can only clear multi-selects')

        for option in self.selected_options:
            option.click()

    def options(self, *args, **kwargs):
        """ Gets all the options in the select list
        :rtype: list[nerodia.elements.option.Option]
        """
        return self._element_call(lambda: super(Select, self).options(*args, **kwargs),
                                  self.wait_for_present)

    def includes(self, term):
        """
        Returns True if the select list has at least one option where text or label matches the
        given value
        :param term: string or regex to match against the option
        :rtype: bool
        """
        return self.option(text=term).exists or self.option(label=term).exists

    def select(self, *terms):
        """
        Select the option whose text or label matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param terms: string or regex or list to match against the option
        :return: The text of the option selected. If multiple options match, returns the first match
        :rtype: str
        """
        return [self._select_by(t) for t in self._flatten(terms)][0]

    def select_all(self, *terms):
        """
        Select all the options whose text or label matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param term: string or regex or list to match against the option
        :return: The text of the first option selected.
        :rtype: str
        """
        return [self._select_all_by(t) for t in self._flatten(terms)][0]

    def js_select(self, *terms):
        """
        Uses JavaScript to select the option whose text matches the given string.
        :param term: string or regex or list to match against the option
        """
        return [self._js_select_by(t, 'single') for t in self._flatten(terms)][0]

    def js_select_all(self, *terms):
        """
        Uses JavaScript to select all options whose text matches the given string.
        :param term: string or regex or list to match against the option
        """
        return [self._js_select_by(t, 'multiple') for t in self._flatten(terms)][0]

    def select_value(self, value):
        """
        Selects the option(s) whose value attribute matches the given string
        If this is a multi-select and several options match the value given, all will be selected
        :param value: string or regex to match against the option
        """
        nerodia.logger.deprecate('#select_value', '#select', ids=['select_value'])
        return self._select_by(value)

    def js_select_value(self, value):
        """
        Uses JavaScript to select the option whose value attribute matches the given string
        :param value: string or regex to match against the option
        """
        return self._js_select_by(value, 'single')

    def is_selected(self, term):
        """
        Returns True if any of the selected options' text or label matches the given value
        :param term: string or regex to match against the option
        :rtype: bool
        :raises: UnknownObjectException
        """
        by_text = self.options(text=term)
        if any(option.is_selected for option in by_text):
            return True

        by_label = self.options(label=term)
        if any(option.is_selected for option in by_label):
            return True

        if len(by_text) + len(by_label) != 0:
            return False

        raise UnknownObjectException('Unable to locate option matching {}'.format(term))

    @property
    def value(self):
        """
        Returns the value of the first selected option in the select list
        Returns None if no option is selected
        :rtype: str or None
        """
        selected = self.selected_options
        return selected[0].value if selected else None

    @property
    def text(self):
        # Returns the text of the first selected option in the select list
        # Returns nil if no option is selected
        # :rtype: str or None
        selected = self.selected_options
        return selected[0].text if selected else None

    @property
    def selected_options(self):
        """
        Returns an array of currently selected options
        :rtype: list[nerodia.elements.option.Option]
        """
        return self._element_call(lambda: self._execute_js('selectedOptions', self))

    # private

    def _select_by(self, term):
        found = self._find_options('value', term)
        if found:
            if len(found) > 1:
                nerodia.logger.deprecate('Selecting Multiple Options with #select', '#select_all',
                                         ids=['select_by'])
            return self._select_matching(found)

        raise NoValueFoundException('{} not found in select list'.format(term))

    def _js_select_by(self, term, number):
        if isinstance(term, Pattern):
            js_rx = term.pattern
            js_rx = js_rx.replace('\\A', '^', 1)
            js_rx = js_rx.replace('\\Z', '$', 1)
            js_rx = js_rx.replace('\\z', '$', 1)
            js_rx = re.sub(r'\(\?#.+\)', '', js_rx)
            js_rx = re.sub(r'\(\?-\w+:', '(', js_rx)
        elif type(term) in [six.text_type, six.binary_type]:
            js_rx = '^{}$'.format(term)
        else:
            raise TypeError('expected String or Regexp, got {}'.format(term))

        for way in ['text', 'label', 'value']:
            self._element_call(lambda: self._execute_js('selectOptions{}'.format(way.capitalize()),
                                                        self, js_rx, str(number)))
            if self._is_matching_option(way, term):
                return self.selected_options[0].text

        raise NoValueFoundException('{} not found in select list'.format(term))

    def _is_matching_option(self, how, what):
        for opt in self.selected_options:
            value = getattr(opt, how)
            if (type(what) in [six.text_type, six.binary_type] and value == what) or \
                    (type(what) not in [six.text_type, six.binary_type] and re.search(what, value)):
                if opt.enabled:
                    return True
                raise ObjectDisabledException('option matching {} by {} on {}'
                                              ' is disabled'.format(what, how, self))

    def _select_all_by(self, term):
        if not self.multiple:
            raise Error('you can only use #select_all on multi-selects')

        found = self._find_options('text', term)
        if found:
            return self._select_matching(found)

        raise NoValueFoundException('{} not found in select list'.format(term))

    def _find_options(self, how, term):
        types = [six.text_type, six.binary_type, int, Pattern]
        found = []

        def func(sel):
            if type(term) in types:
                collection = sel.options(value=term) if how == 'value' else []
                if not list(collection):
                    collection = sel.options(text=term)
                if not list(collection):
                    collection = sel.options(label=term)
                if collection:
                    found.append(collection)
                    return False
                else:
                    return not found and nerodia.relaxed_locate
            else:
                raise TypeError('expected {!r}, got {}:{}'.format(types, term, term.__class__))

        try:
            Wait.until_not(func, object=self)
            if found:
                return found[0]
            raise NoValueFoundException('{} not found in select list'.format(term))
        except TimeoutError:
            raise NoValueFoundException('{} not found in select list'.format(term))

    def _select_matching(self, elements):
        if not self.multiple:
            elements = elements[:1]
        for element in elements:
            if not element.is_selected:
                element.click()
        return elements[0].text if elements[0].exist else ''

    def _matches_regexp(self, how, element, exp):
        if how == 'text':
            return (re.search(exp, self.el.text) or re.search(exp, self.el.label)) \
                is not None
        elif how == 'value':
            return re.search(exp, element.value) is not None
        else:
            raise Error('unknown how: {}'.format(how))

    @staticmethod
    def _flatten(term):
        for x in term:
            if isinstance(x, list):
                for y in x:
                    yield y
            else:
                yield x



import six

from .html_elements import InputCollection
from ..elements.input import Input
from ..exception import ObjectDisabledException
from ..meta_elements import MetaHTMLElement


@six.add_metaclass(MetaHTMLElement)
class Hidden(Input):
    @property
    def visible(self):
        return False

    def click(self):
        raise ObjectDisabledException('click is not available on the hidden field element')


@six.add_metaclass(MetaHTMLElement)
class HiddenCollection(InputCollection):
    pass



# Autogenerated from SVG specification. Edits may be lost.

svgelement = [
    (str, 'classname', 'className'),
    (str, 'dataset', 'dataset'),
    (str, 'ownersvgelement', 'ownerSVGElement'),
    (str, 'viewportelement', 'viewportElement'),
    (int, 'tabindex', 'tabIndex'),
    (str, 'correspondingelement', 'correspondingElement'),
    (str, 'correspondinguseelement', 'correspondingUseElement')
]

SVGElementCollection = []

view = [
    (str, 'viewbox', 'viewBox'),
    (str, 'preserveaspectratio', 'preserveAspectRatio'),
    (int, 'zoomandpan', 'zoomAndPan')
]

ViewCollection = []

cursor = [
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (str, 'href', 'href')
]

CursorCollection = []

Hatchpath = []

HatchpathCollection = []

Hatch = []

HatchCollection = []

pattern = [
    (str, 'patternunits', 'patternUnits'),
    (str, 'patterncontentunits', 'patternContentUnits'),
    (str, 'patterntransform', 'patternTransform'),
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (int, 'width', 'width'),
    (int, 'height', 'height'),
    (str, 'viewbox', 'viewBox'),
    (str, 'preserveaspectratio', 'preserveAspectRatio'),
    (str, 'href', 'href')
]

PatternCollection = []

stop = [
    (str, 'offset', 'offset')
]

StopCollection = []

Meshpatch = []

MeshpatchCollection = []

Meshrow = []

MeshrowCollection = []

gradient = [
    (str, 'gradientunits', 'gradientUnits'),
    (str, 'gradienttransform', 'gradientTransform'),
    (str, 'spreadmethod', 'spreadMethod'),
    (str, 'href', 'href')
]

GradientCollection = []

MeshGradient = []

MeshGradientCollection = []

radialgradient = [
    (int, 'cx', 'cx'),
    (int, 'cy', 'cy'),
    (int, 'r', 'r'),
    (int, 'fx', 'fx'),
    (int, 'fy', 'fy'),
    (int, 'fr', 'fr')
]

RadialGradientCollection = []

lineargradient = [
    (int, 'x1', 'x1'),
    (int, 'y1', 'y1'),
    (int, 'x2', 'x2'),
    (int, 'y2', 'y2')
]

LinearGradientCollection = []

Solidcolor = []

SolidcolorCollection = []

marker = [
    (int, 'refx', 'refX'),
    (int, 'refy', 'refY'),
    (str, 'markerunits', 'markerUnits'),
    (int, 'markerwidth', 'markerWidth'),
    (int, 'markerheight', 'markerHeight'),
    (str, 'orienttype', 'orientType'),
    (str, 'orientangle', 'orientAngle'),
    (str, 'orient', 'orient'),
    (str, 'viewbox', 'viewBox'),
    (str, 'preserveaspectratio', 'preserveAspectRatio')
]

MarkerCollection = []

Metadata = []

MetadataCollection = []

Desc = []

DescCollection = []

graphics = [
    (str, 'transform', 'transform'),
    (str, 'requiredextensions', 'requiredExtensions'),
    (str, 'systemlanguage', 'systemLanguage')
]

GraphicsCollection = []

foreignobject = [
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (int, 'width', 'width'),
    (int, 'height', 'height')
]

ForeignObjectCollection = []

textcontent = [
    (int, 'textlength', 'textLength'),
    (str, 'lengthadjust', 'lengthAdjust')
]

TextContentCollection = []

textpath = [
    (int, 'startoffset', 'startOffset'),
    (str, 'method', 'method'),
    (str, 'spacing', 'spacing'),
    (str, 'href', 'href')
]

TextPathCollection = []

textpositioning = [
    (str, 'x', 'x'),
    (str, 'y', 'y'),
    (str, 'dx', 'dx'),
    (str, 'dy', 'dy'),
    (str, 'rotate', 'rotate')
]

TextPositioningCollection = []

TSpan = []

TSpanCollection = []

Switch = []

SwitchCollection = []

use = [
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (int, 'width', 'width'),
    (int, 'height', 'height'),
    (str, 'instanceroot', 'instanceRoot'),
    (str, 'animatedinstanceroot', 'animatedInstanceRoot'),
    (str, 'href', 'href')
]

UseCollection = []

symbol = [
    (str, 'viewbox', 'viewBox'),
    (str, 'preserveaspectratio', 'preserveAspectRatio')
]

SymbolCollection = []

Defs = []

DefsCollection = []

G = []

GCollection = []

svg = [
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (int, 'width', 'width'),
    (int, 'height', 'height'),
    (float, 'currentscale', 'currentScale'),
    (str, 'currenttranslate', 'currentTranslate'),
    (str, 'viewbox', 'viewBox'),
    (str, 'preserveaspectratio', 'preserveAspectRatio'),
    (int, 'zoomandpan', 'zoomAndPan')
]

SVGCollection = []

geometry = [
    (str, 'pathlength', 'pathLength')
]

GeometryCollection = []

polygon = [
    (str, 'points', 'points'),
    (str, 'animatedpoints', 'animatedPoints')
]

PolygonCollection = []

polyline = [
    (str, 'points', 'points'),
    (str, 'animatedpoints', 'animatedPoints')
]

PolylineCollection = []

mesh = [
    (str, 'href', 'href')
]

MeshCollection = []

line = [
    (int, 'x1', 'x1'),
    (int, 'y1', 'y1'),
    (int, 'x2', 'x2'),
    (int, 'y2', 'y2')
]

LineCollection = []

ellipse = [
    (int, 'cx', 'cx'),
    (int, 'cy', 'cy'),
    (int, 'rx', 'rx'),
    (int, 'ry', 'ry')
]

EllipseCollection = []

circle = [
    (int, 'cx', 'cx'),
    (int, 'cy', 'cy'),
    (int, 'r', 'r')
]

CircleCollection = []

rect = [
    (int, 'x', 'x'),
    (int, 'y', 'y'),
    (int, 'width', 'width'),
    (int, 'height', 'height'),
    (int, 'rx', 'rx'),
    (int, 'ry', 'ry')
]

RectCollection = []

Path = []

PathCollection = []



from time import sleep, time


class Timer(object):
    def __init__(self, timeout=None):
        self.end_time = self._current_time() + timeout if timeout else None

    def wait(self, timeout, method, interval=0, object=None, expected=True):
        """
        Executes the given method until it returns True or exceeds the timeout
        :param timeout: time in seconds to timeout after
        :param method: method to call (typically a lambda)
        :param interval: interval to wait between calls
        :param obect: object to call in the method
        :param expected: expected outcome
        :type expected: bool
        :rtype: bool
        """
        end_time = self.end_time or self._current_time() + timeout
        run_once = False
        while True:
            if self._current_time() > end_time and run_once:
                break
            run_once = True
            result = method(object) if object else method()
            if result == expected:
                return True
            sleep(interval)
        return False

    @property
    def remaining_time(self):
        return self.end_time - self._current_time() if self.end_time else 0

    def reset(self):
        self.end_time = None

    @property
    def locked(self):
        return self.end_time is not None

    # private

    def _current_time(self):
        return time()
